@@ -1,28 +1,16 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2015-2022 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 #include "color.h"
+#include "bfstd.h"
 #include "bftw.h"
+#include "config.h"
 #include "dir.h"
 #include "dstring.h"
 #include "expr.h"
 #include "fsade.h"
 #include "stat.h"
 #include "trie.h"
-#include "util.h"
 #include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -131,8 +119,7 @@ static int set_color(struct colors *colors, const char *name, char *value) {
 /**
  * Transform a file extension for fast lookups, by reversing and lowercasing it.
  */
-static void extxfrm(char *ext) {
-	size_t len = strlen(ext);
+static void extxfrm(char *ext, size_t len) {
 	for (size_t i = 0; i < len - i; ++i) {
 		char a = ext[i];
 		char b = ext[len - i - 1];
@@ -153,11 +140,19 @@ static void extxfrm(char *ext) {
 	}
 }
 
+/** Maximum supported extension length. */
+#define EXT_MAX 255
+
 /**
  * Set the color for an extension.
  */
-static int set_ext_color(struct colors *colors, char *key, const char *value) {
-	extxfrm(key);
+static int set_ext_color(struct colors *colors, char *key, char *value) {
+	size_t len = dstrlen(key);
+	if (len > EXT_MAX) {
+		return -1;
+	}
+
+	extxfrm(key, len);
 
 	// A later *.x should override any earlier *.x, *.y.x, etc.
 	struct trie_leaf *match;
@@ -168,7 +163,7 @@ static int set_ext_color(struct colors *colors, char *key, const char *value) {
 
 	struct trie_leaf *leaf = trie_insert_str(&colors->ext_colors, key);
 	if (leaf) {
-		leaf->value = (char *)value;
+		leaf->value = value;
 		return 0;
 	} else {
 		return -1;
@@ -179,14 +174,15 @@ static int set_ext_color(struct colors *colors, char *key, const char *value) {
  * Find a color by an extension.
  */
 static const char *get_ext_color(const struct colors *colors, const char *filename) {
-	char *xfrm = strdup(filename);
-	if (!xfrm) {
-		return NULL;
-	}
-	extxfrm(xfrm);
+	size_t name_len = strlen(filename);
+	size_t ext_len = name_len < EXT_MAX ? name_len : EXT_MAX;
+	const char *ext = filename + name_len - ext_len;
+
+	char xfrm[ext_len + 1];
+	memcpy(xfrm, ext, sizeof(xfrm));
+	extxfrm(xfrm, ext_len);
 
 	const struct trie_leaf *leaf = trie_find_prefix(&colors->ext_colors, xfrm);
-	free(xfrm);
 	if (leaf) {
 		return leaf->value;
 	} else {
@@ -486,17 +482,14 @@ struct colors *parse_colors(void) {
 
 void free_colors(struct colors *colors) {
 	if (colors) {
-		struct trie_leaf *leaf;
-		while ((leaf = trie_first_leaf(&colors->ext_colors))) {
+		TRIE_FOR_EACH(&colors->ext_colors, leaf) {
 			dstrfree(leaf->value);
-			trie_remove(&colors->ext_colors, leaf);
 		}
 		trie_destroy(&colors->ext_colors);
 
-		while ((leaf = trie_first_leaf(&colors->names))) {
+		TRIE_FOR_EACH(&colors->names, leaf) {
 			char **field = leaf->value;
 			dstrfree(*field);
-			trie_remove(&colors->names, leaf);
 		}
 		trie_destroy(&colors->names);
 
@@ -803,7 +796,7 @@ static int print_path_colored(CFILE *cfile, const char *path, const struct BFTW
 	if (path == ftwbuf->path) {
 		nameoff = ftwbuf->nameoff;
 	} else {
-		nameoff = xbasename(path) - path;
+		nameoff = xbaseoff(path);
 	}
 
 	if (print_dirs_colored(cfile, path, ftwbuf, flags, nameoff) != 0) {
@@ -872,6 +865,10 @@ static int cbuff(CFILE *cfile, const char *format, ...);
 
 /** Dump a parsed expression tree, for debugging. */
 static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
+	if (!expr) {
+		return dstrcat(&cfile->buffer, "(null)");
+	}
+
 	if (dstrcat(&cfile->buffer, "(") != 0) {
 		return -1;
 	}
@@ -879,7 +876,7 @@ static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
 	const struct bfs_expr *lhs = NULL;
 	const struct bfs_expr *rhs = NULL;
 
-	if (bfs_expr_has_children(expr)) {
+	if (bfs_expr_is_parent(expr)) {
 		lhs = expr->lhs;
 		rhs = expr->rhs;
 
@@ -935,6 +932,7 @@ static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
 	return 0;
 }
 
+BFS_FORMATTER(2, 0)
 static int cvbuff(CFILE *cfile, const char *format, va_list args) {
 	const struct colors *colors = cfile->colors;
 	int error = errno;
