@@ -1,18 +1,5 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2015-2022 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 #include "ctx.h"
 #include "color.h"
@@ -23,6 +10,7 @@
 #include "pwcache.h"
 #include "stat.h"
 #include "trie.h"
+#include "xtime.h"
 #include <assert.h>
 #include <errno.h>
 #include <limits.h>
@@ -84,9 +72,7 @@ struct bfs_ctx *bfs_ctx_new(void) {
 	ctx->cerr = NULL;
 
 	ctx->users = NULL;
-	ctx->users_error = 0;
 	ctx->groups = NULL;
-	ctx->groups_error = 0;
 
 	ctx->mtab = NULL;
 	ctx->mtab_error = 0;
@@ -95,45 +81,31 @@ struct bfs_ctx *bfs_ctx_new(void) {
 	ctx->nfiles = 0;
 
 	struct rlimit rl;
-	if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
-		ctx->nofile_soft = rl.rlim_cur;
-		ctx->nofile_hard = rl.rlim_max;
-	} else {
-		ctx->nofile_soft = 1024;
-		ctx->nofile_hard = RLIM_INFINITY;
+	if (getrlimit(RLIMIT_NOFILE, &rl) != 0) {
+		goto fail;
+	}
+	ctx->nofile_soft = rl.rlim_cur;
+	ctx->nofile_hard = rl.rlim_max;
+
+	ctx->users = bfs_users_new();
+	if (!ctx->users) {
+		goto fail;
+	}
+
+	ctx->groups = bfs_groups_new();
+	if (!ctx->groups) {
+		goto fail;
+	}
+
+	if (xgettime(&ctx->now) != 0) {
+		goto fail;
 	}
 
 	return ctx;
-}
 
-const struct bfs_users *bfs_ctx_users(const struct bfs_ctx *ctx) {
-	struct bfs_ctx *mut = (struct bfs_ctx *)ctx;
-
-	if (mut->users_error) {
-		errno = mut->users_error;
-	} else if (!mut->users) {
-		mut->users = bfs_users_parse();
-		if (!mut->users) {
-			mut->users_error = errno;
-		}
-	}
-
-	return mut->users;
-}
-
-const struct bfs_groups *bfs_ctx_groups(const struct bfs_ctx *ctx) {
-	struct bfs_ctx *mut = (struct bfs_ctx *)ctx;
-
-	if (mut->groups_error) {
-		errno = mut->groups_error;
-	} else if (!mut->groups) {
-		mut->groups = bfs_groups_parse();
-		if (!mut->groups) {
-			mut->groups_error = errno;
-		}
-	}
-
-	return mut->groups;
+fail:
+	bfs_ctx_free(ctx);
+	return NULL;
 }
 
 const struct bfs_mtab *bfs_ctx_mtab(const struct bfs_ctx *ctx) {
@@ -159,6 +131,8 @@ struct bfs_ctx_file {
 	CFILE *cfile;
 	/** The path to the file (for diagnostics). */
 	const char *path;
+	/** Remembers I/O errors, to propagate them to the exit status. */
+	int error;
 };
 
 CFILE *bfs_ctx_dedup(struct bfs_ctx *ctx, CFILE *cfile, const char *path) {
@@ -189,6 +163,7 @@ CFILE *bfs_ctx_dedup(struct bfs_ctx *ctx, CFILE *cfile, const char *path) {
 
 	ctx_file->cfile = cfile;
 	ctx_file->path = path;
+	ctx_file->error = 0;
 
 	if (cfile != ctx->cout && cfile != ctx->cerr) {
 		++ctx->nfiles;
@@ -202,10 +177,29 @@ void bfs_ctx_flush(const struct bfs_ctx *ctx) {
 	// - the user sees everything relevant before an -ok[dir] prompt
 	// - output from commands is interleaved consistently with bfs
 	// - executed commands can rely on I/O from other bfs actions
-	//
-	// We do not check errors here, but they will be caught at cleanup time
-	// with ferror().
-	fflush(NULL);
+	TRIE_FOR_EACH(&ctx->files, leaf) {
+		struct bfs_ctx_file *ctx_file = leaf->value;
+		CFILE *cfile = ctx_file->cfile;
+		if (fflush(cfile->file) == 0) {
+			continue;
+		}
+
+		ctx_file->error = errno;
+		clearerr(cfile->file);
+
+		const char *path = ctx_file->path;
+		if (path) {
+			bfs_error(ctx, "'%s': %m.\n", path);
+		} else if (cfile == ctx->cout) {
+			bfs_error(ctx, "(standard output): %m.\n");
+		}
+
+	}
+
+	// Flush the user/group caches, in case the executed command edits the
+	// user/group tables
+	bfs_users_flush(ctx->users);
+	bfs_groups_flush(ctx->groups);
 }
 
 /** Flush a file and report any errors. */
@@ -270,10 +264,14 @@ int bfs_ctx_free(struct bfs_ctx *ctx) {
 		bfs_groups_free(ctx->groups);
 		bfs_users_free(ctx->users);
 
-		struct trie_leaf *leaf;
-		while ((leaf = trie_first_leaf(&ctx->files))) {
+		TRIE_FOR_EACH(&ctx->files, leaf) {
 			struct bfs_ctx_file *ctx_file = leaf->value;
 
+			if (ctx_file->error) {
+				// An error was previously reported during bfs_ctx_flush()
+				ret = -1;
+			}
+
 			if (bfs_ctx_fclose(ctx, ctx_file) != 0) {
 				if (cerr) {
 					bfs_error(ctx, "'%s': %m.\n", ctx_file->path);
@@ -282,13 +280,12 @@ int bfs_ctx_free(struct bfs_ctx *ctx) {
 			}
 
 			free(ctx_file);
-			trie_remove(&ctx->files, leaf);
 		}
 		trie_destroy(&ctx->files);
 
 		if (cout && bfs_ctx_fflush(cout) != 0) {
 			if (cerr) {
-				bfs_error(ctx, "standard output: %m.\n");
+				bfs_error(ctx, "(standard output): %m.\n");
 			}
 			ret = -1;
 		}
