@@ -1,24 +1,12 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2019-2021 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 #include "fsade.h"
+#include "config.h"
+#include "bfstd.h"
 #include "bftw.h"
 #include "dir.h"
 #include "dstring.h"
-#include "util.h"
 #include <errno.h>
 #include <fcntl.h>
 #include <stddef.h>
@@ -32,9 +20,9 @@
 #	include <sys/capability.h>
 #endif
 
-#if BFS_HAS_SYS_EXTATTR
+#if BFS_USE_SYS_EXTATTR_H
 #	include <sys/extattr.h>
-#elif BFS_HAS_SYS_XATTR
+#elif BFS_USE_SYS_XATTR_H
 #	include <sys/xattr.h>
 #endif
 
@@ -172,7 +160,7 @@ static int bfs_check_acl_type(acl_t acl, acl_type_t type) {
 #if __FreeBSD__
 	int trivial;
 
-#if BFS_HAS_FEATURE(memory_sanitizer, false)
+#if __has_feature(memory_sanitizer)
         // msan seems to be missing an interceptor for acl_is_trivial_np()
         trivial = 0;
 #endif
@@ -205,7 +193,6 @@ int bfs_check_acl(const struct BFTW *ftwbuf) {
 		ACL_TYPE_NFS4,
 #endif
 	};
-	static const size_t n_acl_types = sizeof(acl_types)/sizeof(acl_types[0]);
 
 	if (ftwbuf->type == BFS_LNK) {
 		return 0;
@@ -214,7 +201,7 @@ int bfs_check_acl(const struct BFTW *ftwbuf) {
 	const char *path = fake_at(ftwbuf);
 
 	int ret = -1, error = 0;
-	for (size_t i = 0; i < n_acl_types && ret <= 0; ++i) {
+	for (size_t i = 0; i < BFS_COUNTOF(acl_types) && ret <= 0; ++i) {
 		acl_type_t type = acl_types[i];
 
 		if (type == ACL_TYPE_DEFAULT && ftwbuf->type != BFS_DIR) {
@@ -303,7 +290,7 @@ int bfs_check_xattrs(const struct BFTW *ftwbuf) {
 	const char *path = fake_at(ftwbuf);
 	ssize_t len;
 
-#if BFS_HAS_SYS_EXTATTR
+#if BFS_USE_SYS_EXTATTR_H
 	ssize_t (*extattr_list)(const char *, int, void*, size_t) =
 		ftwbuf->type == BFS_LNK ? extattr_list_link : extattr_list_file;
 
@@ -342,7 +329,7 @@ int bfs_check_xattr_named(const struct BFTW *ftwbuf, const char *name) {
 	const char *path = fake_at(ftwbuf);
 	ssize_t len;
 
-#if BFS_HAS_SYS_EXTATTR
+#if BFS_USE_SYS_EXTATTR_H
 	ssize_t (*extattr_get)(const char *, int, const char *, void*, size_t) =
 		ftwbuf->type == BFS_LNK ? extattr_get_link : extattr_get_file;
 
