@@ -1,18 +1,5 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2017-2022 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 /**
  * The expression optimizer.  Different optimization levels are supported:
@@ -41,17 +28,18 @@
 
 #include "opt.h"
 #include "color.h"
+#include "config.h"
 #include "ctx.h"
 #include "diag.h"
 #include "eval.h"
+#include "exec.h"
 #include "expr.h"
 #include "pwcache.h"
-#include "util.h"
 #include <assert.h>
+#include <errno.h>
 #include <limits.h>
 #include <stdarg.h>
 #include <stdbool.h>
-#include <stdint.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -340,11 +328,18 @@ static void opt_warning(const struct opt_state *state, const struct bfs_expr *ex
 	if (bfs_expr_warning(state->ctx, expr)) {
 		va_list args;
 		va_start(args, format);
-		bfs_warning(state->ctx, format, args);
+		bfs_vwarning(state->ctx, format, args);
 		va_end(args);
 	}
 }
 
+/** Create a constant expression. */
+static struct bfs_expr *opt_const(bool value) {
+	static bfs_eval_fn *fns[] = {eval_false, eval_true};
+	static char *fake_args[] = {"-false", "-true"};
+	return bfs_expr_new(fns[value], 1, &fake_args[value]);
+}
+
 /** Extract a child expression, freeing the outer expression. */
 static struct bfs_expr *extract_child_expr(struct bfs_expr *expr, struct bfs_expr **child) {
 	struct bfs_expr *ret = *child;
@@ -367,10 +362,6 @@ static struct bfs_expr *negate_expr(struct bfs_expr *rhs, char **argv) {
 		return NULL;
 	}
 
-	if (argv == &fake_not_arg) {
-		expr->synthetic = true;
-	}
-
 	expr->lhs = NULL;
 	expr->rhs = rhs;
 	return expr;
@@ -405,7 +396,6 @@ static struct bfs_expr *de_morgan(const struct opt_state *state, struct bfs_expr
 		expr->eval_fn = eval_and;
 		expr->argv = &fake_and_arg;
 	}
-	expr->synthetic = true;
 
 	expr->lhs = negate_expr(expr->lhs, argv);
 	expr->rhs = negate_expr(expr->rhs, argv);
@@ -463,14 +453,11 @@ static struct bfs_expr *optimize_not_expr(const struct opt_state *state, struct
 
 	int optlevel = state->ctx->optlevel;
 	if (optlevel >= 1) {
-		if (rhs == &bfs_true) {
-			opt_debug(state, 1, "constant propagation: %pe <==> %pe\n", expr, &bfs_false);
+		if (rhs->eval_fn == eval_true || rhs->eval_fn == eval_false) {
+			struct bfs_expr *ret = opt_const(rhs->eval_fn == eval_false);
+			opt_debug(state, 1, "constant propagation: %pe <==> %pe\n", expr, ret);
 			bfs_expr_free(expr);
-			return &bfs_false;
-		} else if (rhs == &bfs_false) {
-			opt_debug(state, 1, "constant propagation: %pe <==> %pe\n", expr, &bfs_true);
-			bfs_expr_free(expr);
-			return &bfs_true;
+			return ret;
 		} else if (rhs->eval_fn == eval_not) {
 			opt_debug(state, 1, "double negation: %pe <==> %pe\n", expr, rhs->rhs);
 			return extract_child_expr(expr, &rhs->rhs);
@@ -520,17 +507,17 @@ static struct bfs_expr *optimize_and_expr(const struct opt_state *state, struct
 	const struct bfs_ctx *ctx = state->ctx;
 	int optlevel = ctx->optlevel;
 	if (optlevel >= 1) {
-		if (lhs == &bfs_true) {
+		if (lhs->eval_fn == eval_true) {
 			opt_debug(state, 1, "conjunction elimination: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
-		} else if (rhs == &bfs_true) {
+		} else if (rhs->eval_fn == eval_true) {
 			opt_debug(state, 1, "conjunction elimination: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (lhs->always_false) {
 			opt_debug(state, 1, "short-circuit: %pe <==> %pe\n", expr, lhs);
 			opt_warning(state, expr->rhs, "This expression is unreachable.\n\n");
 			return extract_child_expr(expr, &expr->lhs);
-		} else if (lhs->always_true && rhs == &bfs_false) {
+		} else if (lhs->always_true && rhs->eval_fn == eval_false) {
 			bool debug = opt_debug(state, 1, "strength reduction: %pe <==> ", expr);
 			struct bfs_expr *ret = extract_child_expr(expr, &expr->lhs);
 			ret = negate_expr(ret, &fake_not_arg);
@@ -538,7 +525,7 @@ static struct bfs_expr *optimize_and_expr(const struct opt_state *state, struct
 				cfprintf(ctx->cerr, "%pe\n", ret);
 			}
 			return ret;
-		} else if (optlevel >= 2 && lhs->pure && rhs == &bfs_false) {
+		} else if (optlevel >= 2 && lhs->pure && rhs->eval_fn == eval_false) {
 			opt_debug(state, 2, "purity: %pe <==> %pe\n", expr, rhs);
 			opt_warning(state, expr->lhs, "The result of this expression is ignored.\n\n");
 			return extract_child_expr(expr, &expr->rhs);
@@ -595,13 +582,13 @@ static struct bfs_expr *optimize_or_expr(const struct opt_state *state, struct b
 			opt_debug(state, 1, "short-circuit: %pe <==> %pe\n", expr, lhs);
 			opt_warning(state, expr->rhs, "This expression is unreachable.\n\n");
 			return extract_child_expr(expr, &expr->lhs);
-		} else if (lhs == &bfs_false) {
+		} else if (lhs->eval_fn == eval_false) {
 			opt_debug(state, 1, "disjunctive syllogism: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
-		} else if (rhs == &bfs_false) {
+		} else if (rhs->eval_fn == eval_false) {
 			opt_debug(state, 1, "disjunctive syllogism: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
-		} else if (lhs->always_false && rhs == &bfs_true) {
+		} else if (lhs->always_false && rhs->eval_fn == eval_true) {
 			bool debug = opt_debug(state, 1, "strength reduction: %pe <==> ", expr);
 			struct bfs_expr *ret = extract_child_expr(expr, &expr->lhs);
 			ret = negate_expr(ret, &fake_not_arg);
@@ -609,7 +596,7 @@ static struct bfs_expr *optimize_or_expr(const struct opt_state *state, struct b
 				cfprintf(ctx->cerr, "%pe\n", ret);
 			}
 			return ret;
-		} else if (optlevel >= 2 && lhs->pure && rhs == &bfs_true) {
+		} else if (optlevel >= 2 && lhs->pure && rhs->eval_fn == eval_true) {
 			opt_debug(state, 2, "purity: %pe <==> %pe\n", expr, rhs);
 			opt_warning(state, expr->lhs, "The result of this expression is ignored.\n\n");
 			return extract_child_expr(expr, &expr->rhs);
@@ -673,11 +660,12 @@ static struct bfs_expr *ignore_result(const struct opt_state *state, struct bfs_
 			}
 		}
 
-		if (optlevel >= 2 && expr->pure && expr != &bfs_false) {
-			opt_debug(state, 2, "ignored result: %pe --> %pe\n", expr, &bfs_false);
+		if (optlevel >= 2 && expr->pure && expr->eval_fn != eval_false) {
+			struct bfs_expr *ret = opt_const(false);
+			opt_debug(state, 2, "ignored result: %pe --> %pe\n", expr, ret);
 			opt_warning(state, expr, "The result of this expression is ignored.\n\n");
 			bfs_expr_free(expr);
-			expr = &bfs_false;
+			return ret;
 		}
 	}
 
@@ -699,8 +687,8 @@ static struct bfs_expr *optimize_comma_expr(const struct opt_state *state, struc
 			opt_debug(state, 1, "reachability: %pe <==> %pe\n", expr, lhs);
 			opt_warning(state, expr->rhs, "This expression is unreachable.\n\n");
 			return extract_child_expr(expr, &expr->lhs);
-		} else if ((lhs->always_true && rhs == &bfs_true)
-			   || (lhs->always_false && rhs == &bfs_false)) {
+		} else if ((lhs->always_true && rhs->eval_fn == eval_true)
+			   || (lhs->always_false && rhs->eval_fn == eval_false)) {
 			opt_debug(state, 1, "redundancy elimination: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (optlevel >= 2 && lhs->pure) {
@@ -747,19 +735,6 @@ static void infer_pred_facts(struct opt_state *state, enum pred_type pred) {
 	constrain_pred(&state->facts_when_false.preds[pred], false);
 }
 
-/** Infer data flow facts about an -{execut,read,writ}able expression. */
-static void infer_access_facts(struct opt_state *state, const struct bfs_expr *expr) {
-	if (expr->num & R_OK) {
-		infer_pred_facts(state, READABLE_PRED);
-	}
-	if (expr->num & W_OK) {
-		infer_pred_facts(state, WRITABLE_PRED);
-	}
-	if (expr->num & X_OK) {
-		infer_pred_facts(state, EXECUTABLE_PRED);
-	}
-}
-
 /** Infer data flow facts about an icmp-style ([+-]N) expression. */
 static void infer_icmp_facts(struct opt_state *state, const struct bfs_expr *expr, enum range_type type) {
 	struct range *range_when_true = &state->facts_when_true.ranges[type];
@@ -787,49 +762,455 @@ static void infer_icmp_facts(struct opt_state *state, const struct bfs_expr *exp
 	}
 }
 
-/** Infer data flow facts about a -gid expression. */
-static void infer_gid_facts(struct opt_state *state, const struct bfs_expr *expr) {
-	infer_icmp_facts(state, expr, GID_RANGE);
+/** Optimize -{execut,read,writ}able. */
+static struct bfs_expr *optimize_access(struct opt_state *state, struct bfs_expr *expr) {
+	expr->probability = 1.0;
 
-	const struct bfs_groups *groups = bfs_ctx_groups(state->ctx);
+	if (expr->num & R_OK) {
+		infer_pred_facts(state, READABLE_PRED);
+		expr->probability *= 0.99;
+	}
+
+	if (expr->num & W_OK) {
+		infer_pred_facts(state, WRITABLE_PRED);
+		expr->probability *= 0.8;
+	}
+
+	if (expr->num & X_OK) {
+		infer_pred_facts(state, EXECUTABLE_PRED);
+		expr->probability *= 0.2;
+	}
+
+	return expr;
+}
+
+/** Optimize -empty. */
+static struct bfs_expr *optimize_empty(struct opt_state *state, struct bfs_expr *expr) {
+	if (state->ctx->optlevel >= 4) {
+		// Since -empty attempts to open and read directories, it may
+		// have side effects such as reporting permission errors, and
+		// thus shouldn't be re-ordered without aggressive optimizations
+		expr->pure = true;
+	}
+
+	return expr;
+}
+
+/** Optimize -{exec,ok}{,dir}. */
+static struct bfs_expr *optimize_exec(struct opt_state *state, struct bfs_expr *expr) {
+	if (expr->exec->flags & BFS_EXEC_MULTI) {
+		expr->always_true = true;
+	} else {
+		expr->cost = 1000000.0;
+	}
+
+	return expr;
+}
+
+/** Optimize -name/-lname/-path. */
+static struct bfs_expr *optimize_fnmatch(struct opt_state *state, struct bfs_expr *expr) {
+	if (strchr(expr->argv[1], '*')) {
+		expr->probability = 0.5;
+	} else {
+		expr->probability = 0.1;
+	}
+
+	return expr;
+}
+
+/** Optimize -gid. */
+static struct bfs_expr *optimize_gid(struct opt_state *state, struct bfs_expr *expr) {
 	struct range *range = &state->facts_when_true.ranges[GID_RANGE];
-	if (groups && range->min == range->max) {
+	if (range->min == range->max) {
 		gid_t gid = range->min;
-		bool nogroup = !bfs_getgrgid(groups, gid);
-		constrain_pred(&state->facts_when_true.preds[NOGROUP_PRED], nogroup);
+		bool nogroup = !bfs_getgrgid(state->ctx->groups, gid);
+		if (errno == 0) {
+			constrain_pred(&state->facts_when_true.preds[NOGROUP_PRED], nogroup);
+		}
 	}
+
+	return expr;
 }
 
-/** Infer data flow facts about a -uid expression. */
-static void infer_uid_facts(struct opt_state *state, const struct bfs_expr *expr) {
-	infer_icmp_facts(state, expr, UID_RANGE);
+/** Optimize -inum. */
+static struct bfs_expr *optimize_inum(struct opt_state *state, struct bfs_expr *expr) {
+	struct range *range = &state->facts_when_true.ranges[INUM_RANGE];
+	if (range->min == range->max) {
+		expr->probability = 0.01;
+	} else {
+		expr->probability = 0.5;
+	}
 
-	const struct bfs_users *users = bfs_ctx_users(state->ctx);
+	return expr;
+}
+
+/** Optimize -links. */
+static struct bfs_expr *optimize_links(struct opt_state *state, struct bfs_expr *expr) {
+	struct range *range = &state->facts_when_true.ranges[LINKS_RANGE];
+	if (1 >= range->min && 1 <= range->max) {
+		expr->probability = 0.99;
+	} else {
+		expr->probability = 0.5;
+	}
+
+	return expr;
+}
+
+/** Optimize -uid. */
+static struct bfs_expr *optimize_uid(struct opt_state *state, struct bfs_expr *expr) {
 	struct range *range = &state->facts_when_true.ranges[UID_RANGE];
-	if (users && range->min == range->max) {
+	if (range->min == range->max) {
 		uid_t uid = range->min;
-		bool nouser = !bfs_getpwuid(users, uid);
-		constrain_pred(&state->facts_when_true.preds[NOUSER_PRED], nouser);
+		bool nouser = !bfs_getpwuid(state->ctx->users, uid);
+		if (errno == 0) {
+			constrain_pred(&state->facts_when_true.preds[NOUSER_PRED], nouser);
+		}
 	}
+
+	return expr;
 }
 
-/** Infer data flow facts about a -samefile expression. */
-static void infer_samefile_facts(struct opt_state *state, const struct bfs_expr *expr) {
+/** Optimize -samefile. */
+static struct bfs_expr *optimize_samefile(struct opt_state *state, struct bfs_expr *expr) {
 	struct range *range_when_true = &state->facts_when_true.ranges[INUM_RANGE];
 	constrain_min(range_when_true, expr->ino);
 	constrain_max(range_when_true, expr->ino);
+	return expr;
 }
 
-/** Infer data flow facts about a -type expression. */
-static void infer_type_facts(struct opt_state *state, const struct bfs_expr *expr) {
+/** Optimize -size. */
+static struct bfs_expr *optimize_size(struct opt_state *state, struct bfs_expr *expr) {
+	struct range *range = &state->facts_when_true.ranges[SIZE_RANGE];
+	if (range->min == range->max) {
+		expr->probability = 0.01;
+	} else {
+		expr->probability = 0.5;
+	}
+
+	return expr;
+}
+
+/** Estimate probability for -x?type. */
+static void estimate_type_probability(struct bfs_expr *expr) {
+	unsigned int types = expr->num;
+
+	expr->probability = 0.0;
+	if (types & (1 << BFS_BLK)) {
+		expr->probability += 0.00000721183;
+	}
+	if (types & (1 << BFS_CHR)) {
+		expr->probability += 0.0000499855;
+	}
+	if (types & (1 << BFS_DIR)) {
+		expr->probability += 0.114475;
+	}
+	if (types & (1 << BFS_DOOR)) {
+		expr->probability += 0.000001;
+	}
+	if (types & (1 << BFS_FIFO)) {
+		expr->probability += 0.00000248684;
+	}
+	if (types & (1 << BFS_REG)) {
+		expr->probability += 0.859772;
+	}
+	if (types & (1 << BFS_LNK)) {
+		expr->probability += 0.0256816;
+	}
+	if (types & (1 << BFS_SOCK)) {
+		expr->probability += 0.0000116881;
+	}
+	if (types & (1 << BFS_WHT)) {
+		expr->probability += 0.000001;
+	}
+}
+
+/** Optimize -type. */
+static struct bfs_expr *optimize_type(struct opt_state *state, struct bfs_expr *expr) {
 	state->facts_when_true.types &= expr->num;
 	state->facts_when_false.types &= ~expr->num;
+
+	estimate_type_probability(expr);
+
+	return expr;
 }
 
-/** Infer data flow facts about an -xtype expression. */
-static void infer_xtype_facts(struct opt_state *state, const struct bfs_expr *expr) {
+/** Optimize -xtype. */
+static struct bfs_expr *optimize_xtype(struct opt_state *state, struct bfs_expr *expr) {
+	if (state->ctx->optlevel >= 4) {
+		// Since -xtype dereferences symbolic links, it may have side
+		// effects such as reporting permission errors, and thus
+		// shouldn't be re-ordered without aggressive optimizations
+		expr->pure = true;
+	}
+
 	state->facts_when_true.xtypes &= expr->num;
 	state->facts_when_false.xtypes &= ~expr->num;
+
+	estimate_type_probability(expr);
+
+	return expr;
+}
+
+/**
+ * Table of pure expressions.
+ */
+static bfs_eval_fn *const opt_pure[] = {
+	eval_access,
+	eval_acl,
+	eval_capable,
+	eval_depth,
+	eval_false,
+	eval_flags,
+	eval_fstype,
+	eval_gid,
+	eval_hidden,
+	eval_inum,
+	eval_links,
+	eval_lname,
+	eval_name,
+	eval_newer,
+	eval_nogroup,
+	eval_nouser,
+	eval_path,
+	eval_perm,
+	eval_regex,
+	eval_samefile,
+	eval_size,
+	eval_sparse,
+	eval_time,
+	eval_true,
+	eval_type,
+	eval_uid,
+	eval_used,
+	eval_xattr,
+	eval_xattrname,
+};
+
+/**
+ * Table of always-true expressions.
+ */
+static bfs_eval_fn *const opt_always_true[] = {
+	eval_fls,
+	eval_fprint,
+	eval_fprint0,
+	eval_fprintf,
+	eval_fprintx,
+	eval_prune,
+	eval_true,
+
+	// Non-returning
+	eval_exit,
+	eval_quit,
+};
+
+/**
+ * Table of always-false expressions.
+ */
+static bfs_eval_fn *const opt_always_false[] = {
+	eval_false,
+
+	// Non-returning
+	eval_exit,
+	eval_quit,
+};
+
+#define FAST_COST       40.0
+#define FNMATCH_COST   400.0
+#define STAT_COST     1000.0
+#define PRINT_COST   20000.0
+
+/**
+ * Table of expression costs.
+ */
+static const struct {
+	/** The evaluation function with this cost. */
+	bfs_eval_fn *eval_fn;
+	/** The matching cost. */
+	float cost;
+} opt_costs[] = {
+	{eval_access,    STAT_COST},
+	{eval_acl,       STAT_COST},
+	{eval_capable,   STAT_COST},
+	{eval_empty, 2 * STAT_COST}, // readdir() is worse than stat()
+	{eval_fls,      PRINT_COST},
+	{eval_fprint,   PRINT_COST},
+	{eval_fprint0,  PRINT_COST},
+	{eval_fprintf,  PRINT_COST},
+	{eval_fprintx,  PRINT_COST},
+	{eval_fstype,    STAT_COST},
+	{eval_gid,       STAT_COST},
+	{eval_inum,      STAT_COST},
+	{eval_links,     STAT_COST},
+	{eval_lname,  FNMATCH_COST},
+	{eval_name,   FNMATCH_COST},
+	{eval_newer,     STAT_COST},
+	{eval_nogroup,   STAT_COST},
+	{eval_nouser,    STAT_COST},
+	{eval_path,   FNMATCH_COST},
+	{eval_perm,      STAT_COST},
+	{eval_samefile,  STAT_COST},
+	{eval_size,      STAT_COST},
+	{eval_sparse,    STAT_COST},
+	{eval_time,      STAT_COST},
+	{eval_uid,       STAT_COST},
+	{eval_used,      STAT_COST},
+	{eval_xattr,     STAT_COST},
+	{eval_xattrname, STAT_COST},
+};
+
+/**
+ * Table of expression probabilities.
+ */
+static const struct {
+	/** The evaluation function with this cost. */
+	bfs_eval_fn *eval_fn;
+	/** The matching probability. */
+	float probability;
+} opt_probs[] = {
+	{eval_acl,       0.00002},
+	{eval_capable,   0.000002},
+	{eval_empty,     0.01},
+	{eval_false,     0.0},
+	{eval_hidden,    0.01},
+	{eval_nogroup,   0.01},
+	{eval_nouser,    0.01},
+	{eval_samefile,  0.01},
+	{eval_true,      1.0},
+	{eval_xattr,     0.01},
+	{eval_xattrname, 0.01},
+};
+
+
+/**
+ * Table of simple predicates.
+ */
+static const struct {
+	/** The evaluation function this optimizer applies to. */
+	bfs_eval_fn *eval_fn;
+	/** The corresponding predicate. */
+	enum pred_type pred;
+} opt_preds[] = {
+	{eval_acl,         ACL_PRED},
+	{eval_capable, CAPABLE_PRED},
+	{eval_empty,     EMPTY_PRED},
+	{eval_hidden,   HIDDEN_PRED},
+	{eval_nogroup, NOGROUP_PRED},
+	{eval_nouser,   NOUSER_PRED},
+	{eval_sparse,   SPARSE_PRED},
+	{eval_xattr,     XATTR_PRED},
+};
+
+/**
+ * Table of simple range comparisons.
+ */
+static const struct {
+	/** The evaluation function this optimizer applies to. */
+	bfs_eval_fn *eval_fn;
+	/** The corresponding range. */
+	enum range_type range;
+} opt_ranges[] = {
+	{eval_depth, DEPTH_RANGE},
+	{eval_gid,     GID_RANGE},
+	{eval_inum,   INUM_RANGE},
+	{eval_links, LINKS_RANGE},
+	{eval_size,   SIZE_RANGE},
+	{eval_uid,     UID_RANGE},
+};
+
+/** Signature for custom optimizer functions. */
+typedef struct bfs_expr *bfs_opt_fn(struct opt_state *state, struct bfs_expr *expr);
+
+/** Table of custom optimizer functions. */
+static const struct {
+	/** The evaluation function this optimizer applies to. */
+	bfs_eval_fn *eval_fn;
+	/** The corresponding optimizer function. */
+	bfs_opt_fn *opt_fn;
+} opt_fns[] = {
+	// Primaries
+	{eval_access,   optimize_access},
+	{eval_empty,    optimize_empty},
+	{eval_exec,     optimize_exec},
+	{eval_gid,      optimize_gid},
+	{eval_inum,     optimize_inum},
+	{eval_links,    optimize_links},
+	{eval_lname,    optimize_fnmatch},
+	{eval_name,     optimize_fnmatch},
+	{eval_path,     optimize_fnmatch},
+	{eval_samefile, optimize_samefile},
+	{eval_size,     optimize_size},
+	{eval_type,     optimize_type},
+	{eval_uid,      optimize_uid},
+	{eval_xtype,    optimize_xtype},
+
+	// Operators
+	{eval_and,   optimize_and_expr_recursive},
+	{eval_comma, optimize_comma_expr_recursive},
+	{eval_not,   optimize_not_expr_recursive},
+	{eval_or,    optimize_or_expr_recursive},
+};
+
+/**
+ * Look up the appropriate optimizer for an expression and call it.
+ */
+static struct bfs_expr *optimize_expr_lookup(struct opt_state *state, struct bfs_expr *expr) {
+	for (size_t i = 0; i < BFS_COUNTOF(opt_pure); ++i) {
+		if (opt_pure[i] == expr->eval_fn) {
+			expr->pure = true;
+			break;
+		}
+	}
+
+	for (size_t i = 0; i < BFS_COUNTOF(opt_always_true); ++i) {
+		if (opt_always_true[i] == expr->eval_fn) {
+			expr->always_true = true;
+			break;
+		}
+	}
+
+	for (size_t i = 0; i < BFS_COUNTOF(opt_always_false); ++i) {
+		if (opt_always_false[i] == expr->eval_fn) {
+			expr->always_false = true;
+			break;
+		}
+	}
+
+	expr->cost = FAST_COST;
+	for (size_t i = 0; i < BFS_COUNTOF(opt_costs); ++i) {
+		if (opt_costs[i].eval_fn == expr->eval_fn) {
+			expr->cost = opt_costs[i].cost;
+			break;
+		}
+	}
+
+	for (size_t i = 0; i < BFS_COUNTOF(opt_probs); ++i) {
+		if (opt_probs[i].eval_fn == expr->eval_fn) {
+			expr->probability = opt_probs[i].probability;
+			break;
+		}
+	}
+
+	for (size_t i = 0; i < BFS_COUNTOF(opt_preds); ++i) {
+		if (opt_preds[i].eval_fn == expr->eval_fn) {
+			infer_pred_facts(state, opt_preds[i].pred);
+			break;
+		}
+	}
+
+	for (size_t i = 0; i < BFS_COUNTOF(opt_ranges); ++i) {
+		if (opt_ranges[i].eval_fn == expr->eval_fn) {
+			infer_icmp_facts(state, expr, opt_ranges[i].range);
+			break;
+		}
+	}
+
+	for (size_t i = 0; i < BFS_COUNTOF(opt_fns); ++i) {
+		if (opt_fns[i].eval_fn == expr->eval_fn) {
+			return opt_fns[i].opt_fn(state, expr);
+		}
+	}
+
+	return expr;
 }
 
 static struct bfs_expr *optimize_expr_recursive(struct opt_state *state, struct bfs_expr *expr) {
@@ -839,68 +1220,23 @@ static struct bfs_expr *optimize_expr_recursive(struct opt_state *state, struct
 	state->facts_when_false = state->facts;
 
 	if (optlevel >= 2 && facts_are_impossible(&state->facts)) {
-		opt_debug(state, 2, "reachability: %pe --> %pe\n", expr, &bfs_false);
+		struct bfs_expr *ret = opt_const(false);
+		opt_debug(state, 2, "reachability: %pe --> %pe\n", expr, ret);
 		opt_warning(state, expr, "This expression is unreachable.\n\n");
 		bfs_expr_free(expr);
-		expr = &bfs_false;
-		goto done;
+		return ret;
 	}
 
-	if (!bfs_expr_has_children(expr) && !expr->pure) {
+	if (!bfs_expr_is_parent(expr) && !expr->pure) {
 		facts_union(state->facts_when_impure, state->facts_when_impure, &state->facts);
 	}
 
-	if (expr->eval_fn == eval_access) {
-		infer_access_facts(state, expr);
-	} else if (expr->eval_fn == eval_acl) {
-		infer_pred_facts(state, ACL_PRED);
-	} else if (expr->eval_fn == eval_capable) {
-		infer_pred_facts(state, CAPABLE_PRED);
-	} else if (expr->eval_fn == eval_depth) {
-		infer_icmp_facts(state, expr, DEPTH_RANGE);
-	} else if (expr->eval_fn == eval_empty) {
-		infer_pred_facts(state, EMPTY_PRED);
-	} else if (expr->eval_fn == eval_gid) {
-		infer_gid_facts(state, expr);
-	} else if (expr->eval_fn == eval_hidden) {
-		infer_pred_facts(state, HIDDEN_PRED);
-	} else if (expr->eval_fn == eval_inum) {
-		infer_icmp_facts(state, expr, INUM_RANGE);
-	} else if (expr->eval_fn == eval_links) {
-		infer_icmp_facts(state, expr, LINKS_RANGE);
-	} else if (expr->eval_fn == eval_nogroup) {
-		infer_pred_facts(state, NOGROUP_PRED);
-	} else if (expr->eval_fn == eval_nouser) {
-		infer_pred_facts(state, NOUSER_PRED);
-	} else if (expr->eval_fn == eval_samefile) {
-		infer_samefile_facts(state, expr);
-	} else if (expr->eval_fn == eval_size) {
-		infer_icmp_facts(state, expr, SIZE_RANGE);
-	} else if (expr->eval_fn == eval_sparse) {
-		infer_pred_facts(state, SPARSE_PRED);
-	} else if (expr->eval_fn == eval_type) {
-		infer_type_facts(state, expr);
-	} else if (expr->eval_fn == eval_uid) {
-		infer_uid_facts(state, expr);
-	} else if (expr->eval_fn == eval_xattr) {
-		infer_pred_facts(state, XATTR_PRED);
-	} else if (expr->eval_fn == eval_xtype) {
-		infer_xtype_facts(state, expr);
-	} else if (expr->eval_fn == eval_not) {
-		expr = optimize_not_expr_recursive(state, expr);
-	} else if (expr->eval_fn == eval_and) {
-		expr = optimize_and_expr_recursive(state, expr);
-	} else if (expr->eval_fn == eval_or) {
-		expr = optimize_or_expr_recursive(state, expr);
-	} else if (expr->eval_fn == eval_comma) {
-		expr = optimize_comma_expr_recursive(state, expr);
-	}
-
+	expr = optimize_expr_lookup(state, expr);
 	if (!expr) {
-		goto done;
+		return NULL;
 	}
 
-	if (bfs_expr_has_children(expr)) {
+	if (bfs_expr_is_parent(expr)) {
 		struct bfs_expr *lhs = expr->lhs;
 		struct bfs_expr *rhs = expr->rhs;
 		if (rhs) {
@@ -916,39 +1252,42 @@ static struct bfs_expr *optimize_expr_recursive(struct opt_state *state, struct
 	}
 
 	if (expr->always_true) {
+		expr->probability = 1.0;
 		set_facts_impossible(&state->facts_when_false);
 	}
 	if (expr->always_false) {
+		expr->probability = 0.0;
 		set_facts_impossible(&state->facts_when_true);
 	}
 
-	if (optlevel < 2 || expr == &bfs_true || expr == &bfs_false) {
-		goto done;
+	if (optlevel < 2 || expr->eval_fn == eval_true || expr->eval_fn == eval_false) {
+		return expr;
 	}
 
 	if (facts_are_impossible(&state->facts_when_true)) {
 		if (expr->pure) {
-			opt_debug(state, 2, "data flow: %pe --> %pe\n", expr, &bfs_false);
+			struct bfs_expr *ret = opt_const(false);
 			opt_warning(state, expr, "This expression is always false.\n\n");
+			opt_debug(state, 2, "data flow: %pe --> %pe\n", expr, ret);
 			bfs_expr_free(expr);
-			expr = &bfs_false;
+			return ret;
 		} else {
 			expr->always_false = true;
 			expr->probability = 0.0;
 		}
 	} else if (facts_are_impossible(&state->facts_when_false)) {
 		if (expr->pure) {
-			opt_debug(state, 2, "data flow: %pe --> %pe\n", expr, &bfs_true);
+			struct bfs_expr *ret = opt_const(true);
+			opt_debug(state, 2, "data flow: %pe --> %pe\n", expr, ret);
 			opt_warning(state, expr, "This expression is always true.\n\n");
 			bfs_expr_free(expr);
-			expr = &bfs_true;
+			return ret;
 		} else {
 			expr->always_true = true;
 			expr->probability = 1.0;
 		}
 	}
 
-done:
 	return expr;
 }
 
@@ -978,7 +1317,7 @@ static bool reorder_expr(const struct opt_state *state, struct bfs_expr *expr, f
  *         Whether any subexpression was reordered.
  */
 static bool reorder_expr_recursive(const struct opt_state *state, struct bfs_expr *expr) {
-	if (!bfs_expr_has_children(expr)) {
+	if (!bfs_expr_is_parent(expr)) {
 		return false;
 	}
 
