@@ -1,22 +1,11 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2017-2022 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 #include "printf.h"
+#include "bfstd.h"
 #include "bftw.h"
 #include "color.h"
+#include "config.h"
 #include "ctx.h"
 #include "darray.h"
 #include "diag.h"
@@ -26,7 +15,6 @@
 #include "mtab.h"
 #include "pwcache.h"
 #include "stat.h"
-#include "util.h"
 #include "xtime.h"
 #include <assert.h>
 #include <errno.h>
@@ -58,7 +46,7 @@ struct bfs_printf {
 	/** Character data associated with this directive. */
 	char c;
 	/** Some data used by the directive. */
-	const void *ptr;
+	void *ptr;
 };
 
 /** Print some text as-is. */
@@ -90,6 +78,21 @@ static bool should_color(CFILE *cfile, const struct bfs_printf *directive) {
 	assert(ret >= 0 && (size_t)ret < sizeof(buf));			\
 	(void)ret
 
+/**
+ * Common entry point for fprintf() with a dynamic format string.
+ */
+static int dyn_fprintf(FILE *file, const struct bfs_printf *directive, ...) {
+	va_list args;
+	va_start(args, directive);
+
+	BFS_SUPPRESS("-Wformat-nonliteral");
+	int ret = vfprintf(file, directive->str, args);
+	BFS_UNSUPPRESS();
+
+	va_end(args);
+	return ret;
+}
+
 /** %a, %c, %t: ctime() */
 static int bfs_printf_ctime(CFILE *cfile, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
 	// Not using ctime() itself because GNU find adds nanoseconds
@@ -121,7 +124,7 @@ static int bfs_printf_ctime(CFILE *cfile, const struct bfs_printf *directive, co
 	               (long)ts->tv_nsec,
 	               1900 + tm.tm_year);
 
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %A, %B/%W, %C, %T: strftime() */
@@ -182,14 +185,16 @@ static int bfs_printf_strftime(CFILE *cfile, const struct bfs_printf *directive,
 	// POSIX strftime() features
 	default:
 		format[1] = directive->c;
+		BFS_SUPPRESS("-Wformat-nonliteral");
 		ret = strftime(buf, sizeof(buf), format, &tm);
+		BFS_UNSUPPRESS();
 		break;
 	}
 
 	assert(ret >= 0 && (size_t)ret < sizeof(buf));
 	(void)ret;
 
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %b: blocks */
@@ -201,12 +206,12 @@ static int bfs_printf_b(CFILE *cfile, const struct bfs_printf *directive, const
 
 	uintmax_t blocks = ((uintmax_t)statbuf->blocks*BFS_STAT_BLKSIZE + 511)/512;
 	BFS_PRINTF_BUF(buf, "%ju", blocks);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %d: depth */
 static int bfs_printf_d(CFILE *cfile, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
-	return fprintf(cfile->file, directive->str, (intmax_t)ftwbuf->depth);
+	return dyn_fprintf(cfile->file, directive, (intmax_t)ftwbuf->depth);
 }
 
 /** %D: device */
@@ -217,7 +222,7 @@ static int bfs_printf_D(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)statbuf->dev);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %f: file name */
@@ -225,7 +230,7 @@ static int bfs_printf_f(CFILE *cfile, const struct bfs_printf *directive, const
 	if (should_color(cfile, directive)) {
 		return cfprintf(cfile, "%pF", ftwbuf);
 	} else {
-		return fprintf(cfile->file, directive->str, ftwbuf->path + ftwbuf->nameoff);
+		return dyn_fprintf(cfile->file, directive, ftwbuf->path + ftwbuf->nameoff);
 	}
 }
 
@@ -237,7 +242,11 @@ static int bfs_printf_F(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	const char *type = bfs_fstype(directive->ptr, statbuf);
-	return fprintf(cfile->file, directive->str, type);
+	if (!type) {
+		return -1;
+	}
+
+	return dyn_fprintf(cfile->file, directive, type);
 }
 
 /** %G: gid */
@@ -248,7 +257,7 @@ static int bfs_printf_G(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)statbuf->gid);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %g: group name */
@@ -258,13 +267,13 @@ static int bfs_printf_g(CFILE *cfile, const struct bfs_printf *directive, const
 		return -1;
 	}
 
-	const struct bfs_groups *groups = directive->ptr;
-	const struct group *grp = groups ? bfs_getgrgid(groups, statbuf->gid) : NULL;
+	struct bfs_groups *groups = directive->ptr;
+	const struct group *grp = bfs_getgrgid(groups, statbuf->gid);
 	if (!grp) {
 		return bfs_printf_G(cfile, directive, ftwbuf);
 	}
 
-	return fprintf(cfile->file, directive->str, grp->gr_name);
+	return dyn_fprintf(cfile->file, directive, grp->gr_name);
 }
 
 /** %h: leading directories */
@@ -293,7 +302,7 @@ static int bfs_printf_h(CFILE *cfile, const struct bfs_printf *directive, const
 	if (should_color(cfile, directive)) {
 		ret = cfprintf(cfile, "${di}%s${rs}", buf);
 	} else {
-		ret = fprintf(cfile->file, directive->str, buf);
+		ret = dyn_fprintf(cfile->file, directive, buf);
 	}
 
 	free(copy);
@@ -309,7 +318,7 @@ static int bfs_printf_H(CFILE *cfile, const struct bfs_printf *directive, const
 			return cfprintf(cfile, "${di}%s${rs}", ftwbuf->root);
 		}
 	} else {
-		return fprintf(cfile->file, directive->str, ftwbuf->root);
+		return dyn_fprintf(cfile->file, directive, ftwbuf->root);
 	}
 }
 
@@ -321,7 +330,7 @@ static int bfs_printf_i(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)statbuf->ino);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %k: 1K blocks */
@@ -333,7 +342,7 @@ static int bfs_printf_k(CFILE *cfile, const struct bfs_printf *directive, const
 
 	uintmax_t blocks = ((uintmax_t)statbuf->blocks*BFS_STAT_BLKSIZE + 1023)/1024;
 	BFS_PRINTF_BUF(buf, "%ju", blocks);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %l: link target */
@@ -355,7 +364,7 @@ static int bfs_printf_l(CFILE *cfile, const struct bfs_printf *directive, const
 		}
 	}
 
-	int ret = fprintf(cfile->file, directive->str, target);
+	int ret = dyn_fprintf(cfile->file, directive, target);
 	free(buf);
 	return ret;
 }
@@ -367,7 +376,7 @@ static int bfs_printf_m(CFILE *cfile, const struct bfs_printf *directive, const
 		return -1;
 	}
 
-	return fprintf(cfile->file, directive->str, (unsigned int)(statbuf->mode & 07777));
+	return dyn_fprintf(cfile->file, directive, (unsigned int)(statbuf->mode & 07777));
 }
 
 /** %M: symbolic mode */
@@ -379,7 +388,7 @@ static int bfs_printf_M(CFILE *cfile, const struct bfs_printf *directive, const
 
 	char buf[11];
 	xstrmode(statbuf->mode, buf);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %n: link count */
@@ -390,7 +399,7 @@ static int bfs_printf_n(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)statbuf->nlink);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %p: full path */
@@ -398,7 +407,7 @@ static int bfs_printf_p(CFILE *cfile, const struct bfs_printf *directive, const
 	if (should_color(cfile, directive)) {
 		return cfprintf(cfile, "%pP", ftwbuf);
 	} else {
-		return fprintf(cfile->file, directive->str, ftwbuf->path);
+		return dyn_fprintf(cfile->file, directive, ftwbuf->path);
 	}
 }
 
@@ -419,7 +428,7 @@ static int bfs_printf_P(CFILE *cfile, const struct bfs_printf *directive, const
 		copybuf.nameoff -= offset;
 		return cfprintf(cfile, "%pP", &copybuf);
 	} else {
-		return fprintf(cfile->file, directive->str, ftwbuf->path + offset);
+		return dyn_fprintf(cfile->file, directive, ftwbuf->path + offset);
 	}
 }
 
@@ -431,7 +440,7 @@ static int bfs_printf_s(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)statbuf->size);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %S: sparseness */
@@ -447,7 +456,7 @@ static int bfs_printf_S(CFILE *cfile, const struct bfs_printf *directive, const
 	} else {
 		sparsity = (double)BFS_STAT_BLKSIZE*statbuf->blocks/statbuf->size;
 	}
-	return fprintf(cfile->file, directive->str, sparsity);
+	return dyn_fprintf(cfile->file, directive, sparsity);
 }
 
 /** %U: uid */
@@ -458,7 +467,7 @@ static int bfs_printf_U(CFILE *cfile, const struct bfs_printf *directive, const
 	}
 
 	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)statbuf->uid);
-	return fprintf(cfile->file, directive->str, buf);
+	return dyn_fprintf(cfile->file, directive, buf);
 }
 
 /** %u: user name */
@@ -468,13 +477,13 @@ static int bfs_printf_u(CFILE *cfile, const struct bfs_printf *directive, const
 		return -1;
 	}
 
-	const struct bfs_users *users = directive->ptr;
-	const struct passwd *pwd = users ? bfs_getpwuid(users, statbuf->uid) : NULL;
+	struct bfs_users *users = directive->ptr;
+	const struct passwd *pwd = bfs_getpwuid(users, statbuf->uid);
 	if (!pwd) {
 		return bfs_printf_U(cfile, directive, ftwbuf);
 	}
 
-	return fprintf(cfile->file, directive->str, pwd->pw_name);
+	return dyn_fprintf(cfile->file, directive, pwd->pw_name);
 }
 
 static const char *bfs_printf_type(enum bfs_type type) {
@@ -503,7 +512,7 @@ static const char *bfs_printf_type(enum bfs_type type) {
 /** %y: type */
 static int bfs_printf_y(CFILE *cfile, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
 	const char *type = bfs_printf_type(ftwbuf->type);
-	return fprintf(cfile->file, directive->str, type);
+	return dyn_fprintf(cfile->file, directive, type);
 }
 
 /** %Y: target type */
@@ -535,7 +544,7 @@ static int bfs_printf_Y(CFILE *cfile, const struct bfs_printf *directive, const
 		}
 	}
 
-	int ret = fprintf(cfile->file, directive->str, type);
+	int ret = dyn_fprintf(cfile->file, directive, type);
 	if (error != 0) {
 		ret = -1;
 		errno = error;
@@ -732,24 +741,18 @@ int bfs_printf_parse(const struct bfs_ctx *ctx, struct bfs_expr *expr, const cha
 				directive.fn = bfs_printf_f;
 				break;
 			case 'F':
-				directive.ptr = bfs_ctx_mtab(ctx);
+				directive.fn = bfs_printf_F;
+				directive.ptr = (void *)bfs_ctx_mtab(ctx);
 				if (!directive.ptr) {
 					int error = errno;
 					bfs_expr_error(ctx, expr);
 					bfs_error(ctx, "Couldn't parse the mount table: %s.\n", strerror(error));
 					goto directive_error;
 				}
-				directive.fn = bfs_printf_F;
 				break;
 			case 'g':
-				directive.ptr = bfs_ctx_groups(ctx);
-				if (!directive.ptr) {
-					int error = errno;
-					bfs_expr_error(ctx, expr);
-					bfs_error(ctx, "Couldn't parse the group table: %s.\n", strerror(error));
-					goto directive_error;
-				}
 				directive.fn = bfs_printf_g;
+				directive.ptr = ctx->groups;
 				break;
 			case 'G':
 				directive.fn = bfs_printf_G;
@@ -797,14 +800,8 @@ int bfs_printf_parse(const struct bfs_ctx *ctx, struct bfs_expr *expr, const cha
 				directive.stat_field = BFS_STAT_MTIME;
 				break;
 			case 'u':
-				directive.ptr = bfs_ctx_users(ctx);
-				if (!directive.ptr) {
-					int error = errno;
-					bfs_expr_error(ctx, expr);
-					bfs_error(ctx, "Couldn't parse the user table: %s.\n", strerror(error));
-					goto directive_error;
-				}
 				directive.fn = bfs_printf_u;
+				directive.ptr = ctx->users;
 				break;
 			case 'U':
 				directive.fn = bfs_printf_U;
