@@ -1,18 +1,5 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2019 Tavian Barnes <tavianator@tavianator.com>             *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 /**
  * This is an implementation of a "qp trie," as documented at
@@ -22,29 +9,29 @@
  * look like
  *
  *       A    A    A    A
- *     *--->*--->*--->*--->$
- *     |    |    | D    D
- *     |    |    +--->*--->$
- *     |    | B    C    D
- *     |    +--->*--->*--->$
- *     | D    D    A    A
- *     +--->*--->*--->*--->$
- *               | D    D
- *               +--->*--->$
+ *     ●───→●───→●───→●───→○
+ *     │    │    │ D    D
+ *     │    │    └───→●───→○
+ *     │    │ B    C    D
+ *     │    └───→●───→●───→○
+ *     │ D    D    A    A
+ *     └───→●───→●───→●───→○
+ *               │ D    D
+ *               └───→●───→○
  *
  * A compressed (PATRICIA) trie collapses internal nodes that have only a single
  * child, like this:
  *
  *       A    A    AA
- *     *--->*--->*---->$
- *     |    |    | DD
- *     |    |    +---->$
- *     |    | BCD
- *     |    +----->$
- *     | DD    AA
- *     +---->*---->$
- *           | DD
- *           +---->$
+ *     ●───→●───→●────→○
+ *     │    │    │ DD
+ *     │    │    └────→○
+ *     │    │ BCD
+ *     │    └─────→○
+ *     │ DD    AA
+ *     └────→●────→○
+ *           │ DD
+ *           └────→○
  *
  * The nodes can be compressed further by dropping the actual compressed
  * sequences from the nodes, storing it only in the leaves.  This is the
@@ -53,21 +40,39 @@
  * branch on, need to be stored in each node.
  *
  *       A    A    A
- *     0--->1--->2--->AAAA
- *     |    |    | D
- *     |    |    +--->AADD
- *     |    | B
- *     |    +--->ABCD
- *     | D    A
- *     +--->2--->DDAA
- *          | D
- *          +--->DDDD
+ *     0───→1───→2───→AAAA
+ *     │    │    │ D
+ *     │    │    └───→AADD
+ *     │    │ B
+ *     │    └───→ABCD
+ *     │ D    A
+ *     └───→2───→DDAA
+ *          │ D
+ *          └───→DDDD
  *
  * Nodes are represented very compactly.  Rather than a dense array of children,
  * a sparse array of only the non-NULL children directly follows the node in
- * memory.  A bitmap is used to track which children exist; the index of a child
- * i is found by counting the number of bits below bit i that are set.  A tag
- * bit is used to tell pointers to internal nodes apart from pointers to leaves.
+ * memory.  A bitmap is used to track which children exist.
+ *
+ * ┌────────────┐
+ * │       [4] [3] [2][1][0] ←─ children
+ * │        ↓   ↓   ↓  ↓  ↓
+ * │       14  10   6  3  0  ←─ sparse index
+ * │        ↓   ↓   ↓  ↓  ↓
+ * │       0100010001001001  ←─ bitmap
+ * │
+ * │ To convert a sparse index to a dense index, mask off the bits above it, and
+ * │ count the remaining bits.
+ * │
+ * │           10            ←─ sparse index
+ * │            ↓
+ * │       0000001111111111  ←─ mask
+ * │     & 0100010001001001  ←─ bitmap
+ * │       ────────────────
+ * │     = 0000000001001001
+ * │                └──┼──┘
+ * │                  [3]    ←─ dense index
+ * └───────────────────┘
  *
  * This implementation tests a whole nibble (half byte/hex digit) at every
  * branch, so the bitmap takes up 16 bits.  The remainder of a machine word is
@@ -77,7 +82,7 @@
  */
 
 #include "trie.h"
-#include "util.h"
+#include "config.h"
 #include <assert.h>
 #include <limits.h>
 #include <stdbool.h>
@@ -85,8 +90,12 @@
 #include <stdlib.h>
 #include <string.h>
 
-#if CHAR_BIT != 8
-#	error "This trie implementation assumes 8-bit bytes."
+static_assert(CHAR_BIT == 8, "This trie implementation assumes 8-bit bytes.");
+
+#if BFS_TARGET_CLONES && (__i386__ || __x86_64__)
+#	define TARGET_CLONES_POPCNT __attribute__((target_clones("popcnt", "default")))
+#else
+#	define TARGET_CLONES_POPCNT
 #endif
 
 /** Number of bits for the sparse array bitmap, aka the range of a nibble. */
@@ -154,13 +163,18 @@ static uintptr_t trie_encode_node(const struct trie_node *node) {
 
 void trie_init(struct trie *trie) {
 	trie->root = 0;
+	trie->head = NULL;
+	trie->tail = NULL;
+}
+
+/** Check if a number is a power of two. */
+static bool is_power_of_two(size_t n) {
+	return (n & (n - 1)) == 0;
 }
 
 /** Compute the popcount (Hamming weight) of a bitmap. */
 static unsigned int trie_popcount(unsigned int n) {
-#if __POPCNT__
-	// Use the x86 instruction if we have it.  Otherwise, GCC generates a
-	// library call, so use the below implementation instead.
+#if __GNUC__
 	return __builtin_popcount(n);
 #else
 	// See https://en.wikipedia.org/wiki/Hamming_weight#Efficient_implementation
@@ -194,6 +208,7 @@ static unsigned char trie_key_nibble(const void *key, size_t offset) {
  * that case, the first mismatch between the key and the representative will be
  * the depth at which to make a new branch to insert the key.
  */
+TARGET_CLONES_POPCNT
 static struct trie_leaf *trie_representative(const struct trie *trie, const void *key, size_t length) {
 	uintptr_t ptr = trie->root;
 	if (!ptr) {
@@ -219,10 +234,6 @@ static struct trie_leaf *trie_representative(const struct trie *trie, const void
 	return trie_decode_leaf(ptr);
 }
 
-struct trie_leaf *trie_first_leaf(const struct trie *trie) {
-	return trie_representative(trie, NULL, 0);
-}
-
 struct trie_leaf *trie_find_str(const struct trie *trie, const char *key) {
 	return trie_find_mem(trie, key, strlen(key) + 1);
 }
@@ -276,7 +287,8 @@ static bool trie_check_prefix(struct trie_leaf *leaf, size_t skip, const char *k
 	}
 }
 
-struct trie_leaf *trie_find_prefix(const struct trie *trie, const char *key) {
+TARGET_CLONES_POPCNT
+static struct trie_leaf *trie_find_prefix_impl(const struct trie *trie, const char *key) {
 	uintptr_t ptr = trie->root;
 	if (!ptr) {
 		return NULL;
@@ -318,55 +330,125 @@ struct trie_leaf *trie_find_prefix(const struct trie *trie, const char *key) {
 	return best;
 }
 
+struct trie_leaf *trie_find_prefix(const struct trie *trie, const char *key) {
+	return trie_find_prefix_impl(trie, key);
+}
+
 /** Create a new leaf, holding a copy of the given key. */
-static struct trie_leaf *new_trie_leaf(const void *key, size_t length) {
+static struct trie_leaf *trie_leaf_alloc(struct trie *trie, const void *key, size_t length) {
 	struct trie_leaf *leaf = malloc(BFS_FLEX_SIZEOF(struct trie_leaf, key, length));
-	if (leaf) {
-		leaf->value = NULL;
-		leaf->length = length;
-		memcpy(leaf->key, key, length);
+	if (!leaf) {
+		return NULL;
 	}
+
+	leaf->prev = trie->tail;
+	leaf->next = NULL;
+
+	if (leaf->prev) {
+		leaf->prev->next = leaf;
+	} else {
+		trie->head = leaf;
+	}
+
+	trie->tail = leaf;
+
+	leaf->value = NULL;
+	leaf->length = length;
+	memcpy(leaf->key, key, length);
+
 	return leaf;
 }
 
+/** Free a leaf. */
+static void trie_leaf_free(struct trie *trie, struct trie_leaf *leaf) {
+	if (leaf->prev) {
+		leaf->prev->next = leaf->next;
+	} else {
+		trie->head = leaf->next;
+	}
+
+	if (leaf->next) {
+		leaf->next->prev = leaf->prev;
+	} else {
+		trie->tail = leaf->prev;
+	}
+
+	free(leaf);
+}
+
 /** Compute the size of a trie node with a certain number of children. */
 static size_t trie_node_size(unsigned int size) {
 	// Empty nodes aren't supported
 	assert(size > 0);
 	// Node size must be a power of two
-	assert((size & (size - 1)) == 0);
+	assert(is_power_of_two(size));
 
 	return BFS_FLEX_SIZEOF(struct trie_node, children, size);
 }
 
-/** Find the offset of the first nibble that differs between two keys. */
-static size_t trie_key_mismatch(const void *key1, const void *key2, size_t length) {
-	const unsigned char *bytes1 = key1;
-	const unsigned char *bytes2 = key2;
-	size_t i = 0;
-	size_t offset = 0;
-	const size_t chunk = sizeof(size_t);
+#if defined(__BYTE_ORDER__) && __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
+#	define TRIE_BSWAP(n) (n)
+#elif defined(__BYTE_ORDER__) && __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
+#	if __SIZEOF_SIZE_T__ == 8
+#		define TRIE_BSWAP(n) __builtin_bswap64(n)
+#	elif __SIZEOF_SIZE_T__ == 4
+#		define TRIE_BSWAP(n) __builtin_bswap32(n)
+#	endif
+#endif
 
-	for (; i + chunk <= length; i += chunk) {
-		if (memcmp(bytes1 + i, bytes2 + i, chunk) != 0) {
+#ifdef TRIE_BSWAP
+#	if __SIZEOF_SIZE_T__ == __SIZEOF_LONG_LONG__
+#		define TRIE_CTZ(n) __builtin_ctzll(n)
+#	elif __SIZEOF_SIZE_T__ == __SIZEOF_LONG__
+#		define TRIE_CTZ(n) __builtin_ctzl(n)
+#	elif __SIZEOF_SIZE_T__ == __SIZEOF_INT__
+#		define TRIE_CTZ(n) __builtin_ctz(n)
+#	endif
+#endif
+
+/** Find the offset of the first nibble that differs between two keys. */
+static size_t trie_mismatch(const struct trie_leaf *rep, const void *key, size_t length) {
+	if (!rep) {
+		return 0;
+	}
+
+	if (rep->length < length) {
+		length = rep->length;
+	}
+
+	const char *rep_bytes = rep->key;
+	const char *key_bytes = key;
+
+	size_t i = 0;
+	for (size_t chunk = sizeof(chunk); i + chunk <= length; i += chunk) {
+		size_t rep_chunk, key_chunk;
+		memcpy(&rep_chunk, rep_bytes + i, sizeof(rep_chunk));
+		memcpy(&key_chunk, key_bytes + i, sizeof(key_chunk));
+
+		if (rep_chunk != key_chunk) {
+#ifdef TRIE_CTZ
+			size_t diff = TRIE_BSWAP(rep_chunk ^ key_chunk);
+			i *= 2;
+			i += TRIE_CTZ(diff) / 4;
+			return i;
+#else
 			break;
+#endif
 		}
 	}
 
 	for (; i < length; ++i) {
-		unsigned char b1 = bytes1[i], b2 = bytes2[i];
-		if (b1 != b2) {
-			offset = (b1 & 0xF) == (b2 & 0xF);
-			break;
+		unsigned char diff = rep_bytes[i] ^ key_bytes[i];
+		if (diff) {
+			return 2 * i + !(diff & 0xF);
 		}
 	}
 
-	offset |= i << 1;
-	return offset;
+	return 2 * i;
 }
 
 /**
- * Insert a key into a node.  The node must not have a child in that position
+ * Insert a leaf into a node.  The node must not have a child in that position
  * already.  Effectively takes a subtrie like this:
  *
  *     ptr
@@ -383,41 +465,36 @@ static size_t trie_key_mismatch(const void *key1, const void *key2, size_t lengt
  *      v X
  *      *--->...
  *      | Y
- *      +--->key
+ *      +--->leaf
  *      | Z
  *      +--->...
  */
-static struct trie_leaf *trie_node_insert(uintptr_t *ptr, const void *key, size_t length, size_t offset) {
+TARGET_CLONES_POPCNT
+static struct trie_leaf *trie_node_insert(struct trie *trie, uintptr_t *ptr, struct trie_leaf *leaf, unsigned char nibble) {
 	struct trie_node *node = trie_decode_node(*ptr);
 	unsigned int size = trie_popcount(node->bitmap);
 
 	// Double the capacity every power of two
-	if ((size & (size - 1)) == 0) {
-		node = realloc(node, trie_node_size(2*size));
+	if (is_power_of_two(size)) {
+		node = realloc(node, trie_node_size(2 * size));
 		if (!node) {
+			trie_leaf_free(trie, leaf);
 			return NULL;
 		}
 		*ptr = trie_encode_node(node);
 	}
 
-	struct trie_leaf *leaf = new_trie_leaf(key, length);
-	if (!leaf) {
-		return NULL;
-	}
-
-	unsigned char nibble = trie_key_nibble(key, offset);
 	unsigned int bit = 1U << nibble;
 
 	// The child must not already be present
 	assert(!(node->bitmap & bit));
 	node->bitmap |= bit;
 
-	unsigned int index = trie_popcount(node->bitmap & (bit - 1));
-	uintptr_t *child = &node->children[index];
-	if (index < size) {
-		memmove(child + 1, child, (size - index)*sizeof(*child));
+	unsigned int target = trie_popcount(node->bitmap & (bit - 1));
+	for (size_t i = size; i > target; --i) {
+		node->children[i] = node->children[i - 1];
 	}
-	*child = trie_encode_leaf(leaf);
+	node->children[target] = trie_encode_leaf(leaf);
 	return leaf;
 }
 
@@ -482,21 +559,16 @@ static uintptr_t *trie_jump(uintptr_t *ptr, const char *key, size_t *offset) {
  *      v X
  *      *--->*...>--->rep
  *      | Y
- *      +--->key
+ *      +--->leaf
  */
-static struct trie_leaf *trie_split(uintptr_t *ptr, const void *key, size_t length, struct trie_leaf *rep, size_t offset, size_t mismatch) {
-	unsigned char key_nibble = trie_key_nibble(key, mismatch);
+static struct trie_leaf *trie_split(struct trie *trie, uintptr_t *ptr, struct trie_leaf *leaf, struct trie_leaf *rep, size_t offset, size_t mismatch) {
+	unsigned char key_nibble = trie_key_nibble(leaf->key, mismatch);
 	unsigned char rep_nibble = trie_key_nibble(rep->key, mismatch);
 	assert(key_nibble != rep_nibble);
 
 	struct trie_node *node = malloc(trie_node_size(2));
 	if (!node) {
-		return NULL;
-	}
-
-	struct trie_leaf *leaf = new_trie_leaf(key, length);
-	if (!leaf) {
-		free(node);
+		trie_leaf_free(trie, leaf);
 		return NULL;
 	}
 
@@ -520,20 +592,22 @@ struct trie_leaf *trie_insert_str(struct trie *trie, const char *key) {
 	return trie_insert_mem(trie, key, strlen(key) + 1);
 }
 
-struct trie_leaf *trie_insert_mem(struct trie *trie, const void *key, size_t length) {
+TARGET_CLONES_POPCNT
+static struct trie_leaf *trie_insert_mem_impl(struct trie *trie, const void *key, size_t length) {
 	struct trie_leaf *rep = trie_representative(trie, key, length);
-	if (!rep) {
-		struct trie_leaf *leaf = new_trie_leaf(key, length);
-		if (leaf) {
-			trie->root = trie_encode_leaf(leaf);
-		}
-		return leaf;
+	size_t mismatch = trie_mismatch(rep, key, length);
+	if (mismatch >= (length << 1)) {
+		return rep;
 	}
 
-	size_t limit = length < rep->length ? length : rep->length;
-	size_t mismatch = trie_key_mismatch(key, rep->key, limit);
-	if ((mismatch >> 1) >= length) {
-		return rep;
+	struct trie_leaf *leaf = trie_leaf_alloc(trie, key, length);
+	if (!leaf) {
+		return NULL;
+	}
+
+	if (!rep) {
+		trie->root = trie_encode_leaf(leaf);
+		return leaf;
 	}
 
 	size_t offset = 0;
@@ -553,33 +627,38 @@ struct trie_leaf *trie_insert_mem(struct trie *trie, const void *key, size_t len
 			ptr = &node->children[index];
 		} else {
 			assert(offset == mismatch);
-			return trie_node_insert(ptr, key, length, offset);
+			return trie_node_insert(trie, ptr, leaf, nibble);
 		}
 	}
 
 	while (mismatch - offset > OFFSET_MAX) {
 		ptr = trie_jump(ptr, key, &offset);
 		if (!ptr) {
+			trie_leaf_free(trie, leaf);
 			return NULL;
 		}
 	}
 
-	return trie_split(ptr, key, length, rep, offset, mismatch);
+	return trie_split(trie, ptr, leaf, rep, offset, mismatch);
+}
+
+struct trie_leaf *trie_insert_mem(struct trie *trie, const void *key, size_t length) {
+	return trie_insert_mem_impl(trie, key, length);
 }
 
 /** Free a chain of singleton nodes. */
-static void trie_free_singletons(uintptr_t ptr) {
+static void trie_free_singletons(struct trie *trie, uintptr_t ptr) {
 	while (!trie_is_leaf(ptr)) {
 		struct trie_node *node = trie_decode_node(ptr);
 
 		// Make sure the bitmap is a power of two, i.e. it has just one child
-		assert((node->bitmap & (node->bitmap - 1)) == 0);
+		assert(is_power_of_two(node->bitmap));
 
 		ptr = node->children[0];
 		free(node);
 	}
 
-	free(trie_decode_leaf(ptr));
+	trie_leaf_free(trie, trie_decode_leaf(ptr));
 }
 
 /**
@@ -615,7 +694,8 @@ static int trie_collapse_node(uintptr_t *parent, struct trie_node *parent_node,
 	return 0;
 }
 
-void trie_remove(struct trie *trie, struct trie_leaf *leaf) {
+TARGET_CLONES_POPCNT
+static void trie_remove_impl(struct trie *trie, struct trie_leaf *leaf) {
 	uintptr_t *child = &trie->root;
 	uintptr_t *parent = NULL;
 	unsigned int child_bit = 0, child_index = 0;
@@ -632,7 +712,7 @@ void trie_remove(struct trie *trie, struct trie_leaf *leaf) {
 		unsigned int index = trie_popcount(bitmap & (bit - 1));
 
 		// Advance the parent pointer, unless this node had only one child
-		if (bitmap & (bitmap - 1)) {
+		if (!is_power_of_two(bitmap)) {
 			parent = child;
 			child_bit = bit;
 			child_index = index;
@@ -644,14 +724,14 @@ void trie_remove(struct trie *trie, struct trie_leaf *leaf) {
 	assert(trie_decode_leaf(*child) == leaf);
 
 	if (!parent) {
-		trie_free_singletons(trie->root);
+		trie_free_singletons(trie, trie->root);
 		trie->root = 0;
 		return;
 	}
 
 	struct trie_node *node = trie_decode_node(*parent);
 	child = node->children + child_index;
-	trie_free_singletons(*child);
+	trie_free_singletons(trie, *child);
 
 	node->bitmap ^= child_bit;
 	unsigned int parent_size = trie_popcount(node->bitmap);
@@ -664,7 +744,7 @@ void trie_remove(struct trie *trie, struct trie_leaf *leaf) {
 		memmove(child, child + 1, (parent_size - child_index)*sizeof(*child));
 	}
 
-	if ((parent_size & (parent_size - 1)) == 0) {
+	if (is_power_of_two(parent_size)) {
 		node = realloc(node, trie_node_size(parent_size));
 		if (node) {
 			*parent = trie_encode_node(node);
@@ -672,7 +752,12 @@ void trie_remove(struct trie *trie, struct trie_leaf *leaf) {
 	}
 }
 
+void trie_remove(struct trie *trie, struct trie_leaf *leaf) {
+	trie_remove_impl(trie, leaf);
+}
+
 /** Free an encoded pointer to a node. */
+TARGET_CLONES_POPCNT
 static void free_trie_ptr(uintptr_t ptr) {
 	if (trie_is_leaf(ptr)) {
 		free(trie_decode_leaf(ptr));
