@@ -1,29 +1,23 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2018-2022 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 #include "xspawn.h"
-#include "util.h"
+#include "bfstd.h"
+#include "config.h"
 #include <errno.h>
 #include <fcntl.h>
+#include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/resource.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
 
+#if BFS_USE_PATHS_H
+#	include <paths.h>
+#endif
+
 /**
  * Types of spawn actions.
  */
@@ -260,11 +254,15 @@ char *bfs_spawn_resolve(const char *exe) {
 	const char *path = getenv("PATH");
 
 	char *confpath = NULL;
-#ifdef _CS_PATH
 	if (!path) {
+#if defined(_CS_PATH)
 		path = confpath = xconfstr(_CS_PATH);
-	}
+#elif defined(_PATH_DEFPATH)
+		path = _PATH_DEFPATH;
+#else
+		errno = ENOENT;
 #endif
+	}
 	if (!path) {
 		return NULL;
 	}
