@@ -1,25 +1,14 @@
-/****************************************************************************
- * bfs                                                                      *
- * Copyright (C) 2020-2022 Tavian Barnes <tavianator@tavianator.com>        *
- *                                                                          *
- * Permission to use, copy, modify, and/or distribute this software for any *
- * purpose with or without fee is hereby granted.                           *
- *                                                                          *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES *
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         *
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  *
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   *
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    *
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  *
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
- ****************************************************************************/
+// Copyright © Tavian Barnes <tavianator@tavianator.com>
+// SPDX-License-Identifier: 0BSD
 
 #include "xtime.h"
 #include <errno.h>
 #include <limits.h>
 #include <stdbool.h>
 #include <stdlib.h>
+#include <sys/time.h>
 #include <time.h>
+#include <unistd.h>
 
 /** Whether tzset() has been called. */
 static bool tz_is_set = false;
@@ -184,7 +173,7 @@ overflow:
 }
 
 /** Parse some digits from a timestamp. */
-static int parse_timestamp_part(const char **str, size_t n, int *result) {
+static int xgetpart(const char **str, size_t n, int *result) {
 	char buf[n + 1];
 	for (size_t i = 0; i < n; ++i, ++*str) {
 		char c = **str;
@@ -199,7 +188,7 @@ static int parse_timestamp_part(const char **str, size_t n, int *result) {
 	return 0;
 }
 
-int parse_timestamp(const char *str, struct timespec *result) {
+int xgetdate(const char *str, struct timespec *result) {
 	struct tm tm = {
 		.tm_isdst = -1,
 	};
@@ -210,7 +199,7 @@ int parse_timestamp(const char *str, struct timespec *result) {
 	bool local = true;
 
 	// YYYY
-	if (parse_timestamp_part(&str, 4, &tm.tm_year) != 0) {
+	if (xgetpart(&str, 4, &tm.tm_year) != 0) {
 		goto invalid;
 	}
 	tm.tm_year -= 1900;
@@ -219,7 +208,7 @@ int parse_timestamp(const char *str, struct timespec *result) {
 	if (*str == '-') {
 		++str;
 	}
-	if (parse_timestamp_part(&str, 2, &tm.tm_mon) != 0) {
+	if (xgetpart(&str, 2, &tm.tm_mon) != 0) {
 		goto invalid;
 	}
 	tm.tm_mon -= 1;
@@ -228,18 +217,18 @@ int parse_timestamp(const char *str, struct timespec *result) {
 	if (*str == '-') {
 		++str;
 	}
-	if (parse_timestamp_part(&str, 2, &tm.tm_mday) != 0) {
+	if (xgetpart(&str, 2, &tm.tm_mday) != 0) {
 		goto invalid;
 	}
 
 	if (!*str) {
 		goto end;
-	} else if (*str == 'T') {
+	} else if (*str == 'T' || *str == ' ') {
 		++str;
 	}
 
 	// hh
-	if (parse_timestamp_part(&str, 2, &tm.tm_hour) != 0) {
+	if (xgetpart(&str, 2, &tm.tm_hour) != 0) {
 		goto invalid;
 	}
 
@@ -249,7 +238,7 @@ int parse_timestamp(const char *str, struct timespec *result) {
 	} else if (*str == ':') {
 		++str;
 	}
-	if (parse_timestamp_part(&str, 2, &tm.tm_min) != 0) {
+	if (xgetpart(&str, 2, &tm.tm_min) != 0) {
 		goto invalid;
 	}
 
@@ -259,7 +248,7 @@ int parse_timestamp(const char *str, struct timespec *result) {
 	} else if (*str == ':') {
 		++str;
 	}
-	if (parse_timestamp_part(&str, 2, &tm.tm_sec) != 0) {
+	if (xgetpart(&str, 2, &tm.tm_sec) != 0) {
 		goto invalid;
 	}
 
@@ -274,7 +263,7 @@ int parse_timestamp(const char *str, struct timespec *result) {
 		++str;
 
 		// hh
-		if (parse_timestamp_part(&str, 2, &tz_hour) != 0) {
+		if (xgetpart(&str, 2, &tz_hour) != 0) {
 			goto invalid;
 		}
 
@@ -284,7 +273,7 @@ int parse_timestamp(const char *str, struct timespec *result) {
 		} else if (*str == ':') {
 			++str;
 		}
-		if (parse_timestamp_part(&str, 2, &tz_min) != 0) {
+		if (xgetpart(&str, 2, &tz_min) != 0) {
 			goto invalid;
 		}
 	} else {
@@ -321,3 +310,17 @@ invalid:
 error:
 	return -1;
 }
+
+int xgettime(struct timespec *result) {
+#if _POSIX_TIMERS > 0
+	return clock_gettime(CLOCK_REALTIME, result);
+#else
+	struct timeval tv;
+	int ret = gettimeofday(&tv, NULL);
+	if (ret == 0) {
+		result->tv_sec = tv.tv_sec;
+		result->tv_nsec = tv.tv_usec * 1000L;
+	}
+	return ret;
+#endif
+}
