@@ -1,20 +1,7 @@
 #!/usr/bin/env bash
 
-############################################################################
-# bfs                                                                      #
-# Copyright (C) 2015-2022 Tavian Barnes <tavianator@tavianator.com>        #
-#                                                                          #
-# Permission to use, copy, modify, and/or distribute this software for any #
-# purpose with or without fee is hereby granted.                           #
-#                                                                          #
-# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES #
-# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         #
-# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  #
-# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   #
-# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    #
-# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  #
-# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           #
-############################################################################
+# Copyright © Tavian Barnes <tavianator@tavianator.com>
+# SPDX-License-Identifier: 0BSD
 
 set -euP
 umask 022
@@ -22,11 +9,11 @@ umask 022
 export LC_ALL=C
 export TZ=UTC0
 
-export ASAN_OPTIONS="abort_on_error=1"
-export LSAN_OPTIONS="abort_on_error=1"
-export MSAN_OPTIONS="abort_on_error=1"
-export TSAN_OPTIONS="abort_on_error=1"
-export UBSAN_OPTIONS="abort_on_error=1"
+export ASAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
+export LSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
+export MSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
+export TSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
+export UBSAN_OPTIONS="abort_on_error=1:log_to_syslog=0"
 
 export LS_COLORS=""
 unset BFS_COLORS
@@ -53,9 +40,20 @@ fi
 
 UNAME=$(uname)
 
+if [ "$UNAME" = Darwin ]; then
+    # ASan on macOS likes to report
+    #
+    #     malloc: nano zone abandoned due to inability to preallocate reserved vm space.
+    #
+    # to syslog, which as a side effect opens a socket which might take the
+    # place of one of the standard streams if the process is launched with it
+    # closed.  This environment variable avoids the message.
+    export MallocNanoZone=0
+fi
+
 if command -v capsh &>/dev/null; then
     if capsh --has-p=cap_dac_override &>/dev/null || capsh --has-p=cap_dac_read_search &>/dev/null; then
-	if [ -n "$BFS_TRIED_DROP" ]; then
+	if [ -n "${BFS_TRIED_DROP:-}" ]; then
             cat >&2 <<EOF
 ${RED}error:${RST} Failed to drop capabilities.
 EOF
@@ -90,18 +88,15 @@ fi
 function usage() {
     local pad=$(printf "%*s" ${#0} "")
     cat <<EOF
-Usage: ${GRN}$0${RST} [${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}] [${BLU}--posix${RST}] [${BLU}--bsd${RST}] [${BLU}--gnu${RST}] [${BLU}--all${RST}] [${BLU}--sudo${RST}]
-       $pad [${BLU}--stop${RST}] [${BLU}--noclean${RST}] [${BLU}--update${RST}] [${BLU}--verbose${RST}[=${BLD}LEVEL${RST}]] [${BLU}--help${RST}]
-       $pad [${BLD}TEST${RST} [${BLD}TEST${RST} ...]]
+Usage: ${GRN}$0${RST} [${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}] [${BLU}--sudo${RST}[=${BLD}COMMAND${RST}]] [${BLU}--stop${RST}]
+       $pad [${BLU}--noclean${RST}] [${BLU}--update${RST}] [${BLU}--verbose${RST}[=${BLD}LEVEL${RST}]] [${BLU}--help${RST}]
+       $pad [${BLU}--posix${RST}] [${BLU}--bsd${RST}] [${BLU}--gnu${RST}] [${BLU}--all${RST}] [${BLD}TEST${RST} [${BLD}TEST${RST} ...]]
 
   ${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}
       Set the path to the bfs executable to test (default: ${MAG}./bin/bfs${RST})
 
-  ${BLU}--posix${RST}, ${BLU}--bsd${RST}, ${BLU}--gnu${RST}, ${BLU}--all${RST}
-      Choose which test cases to run (default: ${BLU}--all${RST})
-
-  ${BLU}--sudo${RST}
-      Run tests that require root
+  ${BLU}--sudo${RST}[=${BLD}COMMAND${RST}]
+      Run tests that require root using ${GRN}sudo${RST} or the given ${BLD}COMMAND${RST}
 
   ${BLU}--stop${RST}
       Stop when the first error occurs
@@ -126,18 +121,16 @@ Usage: ${GRN}$0${RST} [${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}] [${BLU}--posix
   ${BLU}--help${RST}
       This message
 
+  ${BLU}--posix${RST}, ${BLU}--bsd${RST}, ${BLU}--gnu${RST}, ${BLU}--all${RST}
+      Choose which test cases to run (default: ${BLU}--all${RST})
+
   ${BLD}TEST${RST}
-      Select individual test cases to run (e.g. ${BLD}posix/basic${RST})
+      Select individual test cases to run (e.g. ${BLD}posix/basic${RST}, ${BLD}"*exec*"${RST}, ...)
 EOF
 }
 
-DEFAULT=yes
-POSIX=
-COMMON=
-BSD=
-GNU=
-ALL=
-SUDO=
+PATTERNS=()
+SUDO=()
 STOP=
 CLEAN=yes
 UPDATE=
@@ -145,9 +138,6 @@ VERBOSE_COMMANDS=
 VERBOSE_ERRORS=
 VERBOSE_SKIPPED=
 VERBOSE_TESTS=
-EXPLICIT=
-
-enabled_tests=()
 
 for arg; do
     case "$arg" in
@@ -155,31 +145,22 @@ for arg; do
             BFS="${arg#*=}"
             ;;
         --posix)
-            DEFAULT=
-            POSIX=yes
+            PATTERNS+=("posix/*")
             ;;
         --bsd)
-            DEFAULT=
-            POSIX=yes
-            COMMON=yes
-            BSD=yes
+            PATTERNS+=("posix/*" "common/*" "bsd/*")
             ;;
         --gnu)
-            DEFAULT=
-            POSIX=yes
-            COMMON=yes
-            GNU=yes
+            PATTERNS+=("posix/*" "common/*" "gnu/*")
             ;;
         --all)
-            DEFAULT=
-            POSIX=yes
-            COMMON=yes
-            BSD=yes
-            GNU=yes
-            ALL=yes
+            PATTERNS+=("*")
             ;;
         --sudo)
-            SUDO=yes
+            SUDO=(sudo)
+            ;;
+        --sudo=*)
+            read -a SUDO <<<"${arg#*=}"
             ;;
         --stop)
             STOP=yes
@@ -200,7 +181,6 @@ for arg; do
             VERBOSE_SKIPPED=yes
             ;;
         --verbose=tests)
-            VERBOSE_SKIPPED=yes
             VERBOSE_TESTS=yes
             ;;
         --verbose)
@@ -213,27 +193,17 @@ for arg; do
             usage
             exit 0
             ;;
-        */*)
-            EXPLICIT=yes
-            SUDO=yes
-            enabled_tests+=("$arg")
-            ;;
-        *)
+        -*)
             printf "${RED}error:${RST} Unrecognized option '%s'.\n\n" "$arg" >&2
             usage >&2
             exit 1
             ;;
+        *)
+            PATTERNS+=("$arg")
+            ;;
     esac
 done
 
-if [ "$DEFAULT" ]; then
-    POSIX=yes
-    COMMON=yes
-    BSD=yes
-    GNU=yes
-    ALL=yes
-fi
-
 function _realpath() {
     (
         cd "$(dirname -- "$1")"
@@ -248,6 +218,8 @@ if [ "${BUILDDIR-}" ]; then
 else
     BIN=$(_realpath "$TESTS/../bin")
 fi
+MKSOCK="$BIN/tests/mksock"
+XTOUCH="$BIN/tests/xtouch"
 
 # Try to resolve the path to $BFS before we cd, while also supporting
 # --bfs="./bin/bfs -S ids"
@@ -260,20 +232,41 @@ chown "$(id -u):$(id -g)" "$TMP"
 
 cd "$TESTS"
 
-if [ ! "$EXPLICIT" ]; then
-    [ "$POSIX" ] && enabled_tests+=(posix/*.sh)
-    [ "$COMMON" ] && enabled_tests+=(common/*.sh)
-    [ "$BSD" ] && enabled_tests+=(bsd/*.sh)
-    [ "$GNU" ] && enabled_tests+=(gnu/*.sh)
-    [ "$ALL" ] && enabled_tests+=(bfs/*.sh)
-
-    enabled_tests=("${enabled_tests[@]%.sh}")
+if (( ${#PATTERNS[@]} == 0 )); then
+    PATTERNS=("*")
 fi
 
+TEST_CASES=()
+for TEST in {posix,common,bsd,gnu,bfs}/*.sh; do
+    TEST="${TEST%.sh}"
+    for PATTERN in "${PATTERNS[@]}"; do
+        if [[ $TEST == $PATTERN ]]; then
+            TEST_CASES+=("$TEST")
+            break
+        fi
+    done
+done
+
+if (( ${#TEST_CASES[@]} == 0 )); then
+    printf "${RED}error:${RST} No tests matched" >&2
+    printf " ${BLD}%s${RST}" "${PATTERNS[@]}" >&2
+    printf ".\n\n" >&2
+    usage >&2
+    exit 1
+fi
+
+function bfs_sudo() {
+    if ((${#SUDO[@]})); then
+        "${SUDO[@]}" "$@"
+    else
+        return 1
+    fi
+}
+
 function clean_scratch() {
     if [ -e "$TMP/scratch" ]; then
         # Try to unmount anything left behind
-        if [ "$SUDO" ] && command -v mountpoint &>/dev/null; then
+        if ((${#SUDO[@]})) && command -v mountpoint &>/dev/null; then
             for path in "$TMP"/scratch/*; do
                 if mountpoint -q "$path"; then
                     sudo umount "$path"
@@ -311,63 +304,34 @@ else
     echo "Test files saved to $TMP"
 fi
 
-# Install a file, creating any parent directories
-function installp() {
-    local target="${@: -1}"
-    mkdir -p "${target%/*}"
-    install "$@"
-}
-
-# Prefer GNU touch to work around https://apple.stackexchange.com/a/425730/397839
-if command -v gtouch &>/dev/null; then
-    TOUCH=gtouch
-else
-    TOUCH=touch
-fi
-
-# Like a mythical touch -p
-function touchp() {
-    for arg; do
-        installp -m644 /dev/null "$arg"
-    done
-}
-
 # Creates a simple file+directory structure for tests
 function make_basic() {
-    touchp "$1/a"
-    touchp "$1/b"
-    touchp "$1/c/d"
-    touchp "$1/e/f"
-    mkdir -p "$1/g/h"
-    mkdir -p "$1/i"
-    touchp "$1/j/foo"
-    touchp "$1/k/foo/bar"
-    touchp "$1/l/foo/bar/baz"
+    "$XTOUCH" -p "$1"/{a,b,c/d,e/f,g/h/,i/}
+    "$XTOUCH" -p "$1"/{j/foo,k/foo/bar,l/foo/bar/baz}
     echo baz >"$1/l/foo/bar/baz"
 }
 make_basic "$TMP/basic"
 
 # Creates a file+directory structure with various permissions for tests
 function make_perms() {
-    installp -m000 /dev/null "$1/0"
-    installp -m444 /dev/null "$1/r"
-    installp -m222 /dev/null "$1/w"
-    installp -m644 /dev/null "$1/rw"
-    installp -m555 /dev/null "$1/rx"
-    installp -m311 /dev/null "$1/wx"
-    installp -m755 /dev/null "$1/rwx"
+    "$XTOUCH" -p -M000 "$1/0"
+    "$XTOUCH" -p -M444 "$1/r"
+    "$XTOUCH" -p -M222 "$1/w"
+    "$XTOUCH" -p -M644 "$1/rw"
+    "$XTOUCH" -p -M555 "$1/rx"
+    "$XTOUCH" -p -M311 "$1/wx"
+    "$XTOUCH" -p -M755 "$1/rwx"
 }
 make_perms "$TMP/perms"
 
 # Creates a file+directory structure with various symbolic and hard links
 function make_links() {
-    touchp "$1/file"
+    "$XTOUCH" -p "$1/file"
     ln -s file "$1/symlink"
     ln "$1/file" "$1/hardlink"
     ln -s nowhere "$1/broken"
     ln -s symlink/file "$1/notdir"
-    mkdir -p "$1/deeply/nested/dir"
-    touchp "$1/deeply/nested/file"
+    "$XTOUCH" -p "$1/deeply/nested"/{dir/,file}
     ln -s file "$1/deeply/nested/link"
     ln -s nowhere "$1/deeply/nested/broken"
     ln -s deeply/nested "$1/skip"
@@ -376,7 +340,7 @@ make_links "$TMP/links"
 
 # Creates a file+directory structure with symbolic link loops
 function make_loops() {
-    touchp "$1/file"
+    "$XTOUCH" -p "$1/file"
     ln -s file "$1/symlink"
     ln -s nowhere "$1/broken"
     ln -s symlink/file "$1/notdir"
@@ -389,29 +353,28 @@ make_loops "$TMP/loops"
 
 # Creates a file+directory structure with varying timestamps
 function make_times() {
-    mkdir -p "$1"
-    $TOUCH -t 199112140000 "$1/a"
-    $TOUCH -t 199112140001 "$1/b"
-    $TOUCH -t 199112140002 "$1/c"
+    "$XTOUCH" -p -t "1991-12-14 00:00" "$1/a"
+    "$XTOUCH" -p -t "1991-12-14 00:01" "$1/b"
+    "$XTOUCH" -p -t "1991-12-14 00:02" "$1/c"
     ln -s a "$1/l"
-    $TOUCH -h -t 199112140003 "$1/l"
-    $TOUCH -t 199112140004 "$1"
+    "$XTOUCH" -p -h -t "1991-12-14 00:03" "$1/l"
+    "$XTOUCH" -p -t "1991-12-14 00:04" "$1"
 }
 make_times "$TMP/times"
 
 # Creates a file+directory structure with various weird file/directory names
 function make_weirdnames() {
-    touchp "$1/-/a"
-    touchp "$1/(/b"
-    touchp "$1/(-/c"
-    touchp "$1/!/d"
-    touchp "$1/!-/e"
-    touchp "$1/,/f"
-    touchp "$1/)/g"
-    touchp "$1/.../h"
-    touchp "$1/\\/i"
-    touchp "$1/ /j"
-    touchp "$1/[/k"
+    "$XTOUCH" -p "$1/-/a"
+    "$XTOUCH" -p "$1/(/b"
+    "$XTOUCH" -p "$1/(-/c"
+    "$XTOUCH" -p "$1/!/d"
+    "$XTOUCH" -p "$1/!-/e"
+    "$XTOUCH" -p "$1/,/f"
+    "$XTOUCH" -p "$1/)/g"
+    "$XTOUCH" -p "$1/.../h"
+    "$XTOUCH" -p "$1/\\/i"
+    "$XTOUCH" -p "$1/ /j"
+    "$XTOUCH" -p "$1/[/k"
 }
 make_weirdnames "$TMP/weirdnames"
 
@@ -439,7 +402,7 @@ function make_deep() {
                 cd "$names"
             done
 
-            $TOUCH "$name"
+            "$XTOUCH" "$name"
         )
     done
 }
@@ -447,24 +410,24 @@ make_deep "$TMP/deep"
 
 # Creates a directory structure with many different types, and therefore colors
 function make_rainbow() {
-    touchp "$1/file.txt"
-    touchp "$1/file.dat"
-    touchp "$1/star".{gz,tar,tar.gz}
+    "$XTOUCH" -p "$1/file.txt"
+    "$XTOUCH" -p "$1/file.dat"
+    "$XTOUCH" -p "$1/star".{gz,tar,tar.gz}
     ln -s file.txt "$1/link.txt"
-    touchp "$1/mh1"
+    "$XTOUCH" -p "$1/mh1"
     ln "$1/mh1" "$1/mh2"
     mkfifo "$1/pipe"
     # TODO: block
     ln -s /dev/null "$1/chardev_link"
     ln -s nowhere "$1/broken"
-    "$BIN/tests/mksock" "$1/socket"
-    touchp "$1"/s{u,g,ug}id
+    "$MKSOCK" "$1/socket"
+    "$XTOUCH" -p "$1"/s{u,g,ug}id
     chmod u+s "$1"/su{,g}id
     chmod g+s "$1"/s{u,}gid
     mkdir "$1/ow" "$1"/sticky{,_ow}
     chmod o+w "$1"/*ow
     chmod +t "$1"/sticky*
-    touchp "$1"/exec.sh
+    "$XTOUCH" -p "$1"/exec.sh
     chmod +x "$1"/exec.sh
 }
 make_rainbow "$TMP/rainbow"
@@ -510,20 +473,22 @@ function bfs_verbose() {
 function invoke_bfs() {
     bfs_verbose "$@"
     "${BFS[@]}" "$@"
+    local status="$?"
+
+    # Allow bfs to fail, but not crash
+    if ((status > 125)); then
+        exit "$status"
+    else
+        return "$status"
+    fi
 }
 
-# Expect a command to fail, but not crash
-function fail() {
-    "$@"
-    local STATUS="$?"
-
-    if ((STATUS > 125)); then
-        exit "$STATUS"
-    elif ((STATUS > 0)); then
-        return 0
-    else
-        return 1
-    fi
+function check_exit() {
+    local expected="$1"
+    local actual="0"
+    shift
+    "$@" || actual="$?"
+    ((actual == expected))
 }
 
 # Detect colored diff support
@@ -533,8 +498,6 @@ else
     DIFF="diff"
 fi
 
-# Return value when bfs fails
-EX_BFS=10
 # Return value when a difference is detected
 EX_DIFF=20
 # Return value when a test is skipped
@@ -562,29 +525,23 @@ function bfs_diff() (
     exec 3>&-
 
     "${BFS[@]}" "$@" | sort >"$OUT"
-    local STATUS="${PIPESTATUS[0]}"
+    local status="${PIPESTATUS[0]}"
 
-    diff_output || return $EX_DIFF
-
-    if [ "$STATUS" -eq 0 ]; then
-        return 0
-    else
-        return $EX_BFS
-    fi
+    diff_output || exit $EX_DIFF
+    return "$status"
 )
 
 function skip() {
-    exit $EX_SKIP
-}
-
-function skip_if() {
-    if "$@"; then
-        skip
+    if [ "$VERBOSE_SKIPPED" ]; then
+        caller | {
+            read -r line file
+            printf "${BOL}${CYN}%s skipped!${RST} (%s)\n" "$TEST" "$(awk "NR == $line" "$file")"
+        }
+    elif [ "$VERBOSE_TESTS" ]; then
+        printf "${BOL}${CYN}%s skipped!${RST}\n" "$TEST"
     fi
-}
 
-function skip_unless() {
-    skip_if fail "$@"
+    exit $EX_SKIP
 }
 
 function closefrom() {
@@ -631,7 +588,7 @@ function set_acl() {
 function make_xattrs() {
     clean_scratch
 
-    $TOUCH scratch/{normal,xattr,xattr_2}
+    "$XTOUCH" scratch/{normal,xattr,xattr_2}
     ln -s xattr scratch/link
     ln -s normal scratch/xattr_link
 
@@ -649,10 +606,9 @@ function make_xattrs() {
         *)
             # Linux tmpfs doesn't support the user.* namespace, so we use the security.*
             # namespace, which is writable by root and readable by others
-            [ "$SUDO" ] \
-                && sudo setfattr -n security.bfs_test scratch/xattr \
-                && sudo setfattr -n security.bfs_test_2 scratch/xattr_2 \
-                && sudo setfattr -h -n security.bfs_test scratch/xattr_link
+            bfs_sudo setfattr -n security.bfs_test scratch/xattr \
+                && bfs_sudo setfattr -n security.bfs_test_2 scratch/xattr_2 \
+                && bfs_sudo setfattr -h -n security.bfs_test scratch/xattr_link
             ;;
     esac
 }
@@ -664,9 +620,15 @@ BOL='\n'
 EOL='\n'
 
 function update_eol() {
+    # Bash gets $COLUMNS from stderr, so if it's redirected use tput instead
+    local cols="${COLUMNS-}"
+    if [ -z "$cols" ]; then
+        cols=$(tput cols)
+    fi
+
     # Put the cursor at the last column, then write a space so the next
     # character will wrap
-    EOL="\\033[${COLUMNS}G "
+    EOL="\\033[${cols}G "
 }
 
 if [ "$VERBOSE_TESTS" ]; then
@@ -688,7 +650,7 @@ passed=0
 failed=0
 skipped=0
 
-for TEST in "${enabled_tests[@]}"; do
+for TEST in "${TEST_CASES[@]}"; do
     if [[ -t 1 || "$VERBOSE_TESTS" ]]; then
         printf "${BOL}${YLW}%s${RST}${EOL}" "$TEST"
     else
@@ -699,9 +661,9 @@ for TEST in "${enabled_tests[@]}"; do
     mkdir -p "${OUT%/*}"
 
     if [ "$VERBOSE_ERRORS" ]; then
-        (. "$TESTS/$TEST.sh")
+        (set -e; . "$TESTS/$TEST.sh")
     else
-        (. "$TESTS/$TEST.sh") 2>"$TMP/stderr"
+        (set -e; . "$TESTS/$TEST.sh") 2>"$TMP/$TEST.err"
     fi
     status=$?
 
@@ -709,12 +671,9 @@ for TEST in "${enabled_tests[@]}"; do
         ((++passed))
     elif ((status == EX_SKIP)); then
         ((++skipped))
-        if [ "$VERBOSE_SKIPPED" ]; then
-            printf "${BOL}${CYN}%s skipped!${RST}\n" "$TEST"
-        fi
     else
         ((++failed))
-        [ "$VERBOSE_ERRORS" ] || cat "$TMP/stderr" >&2
+        [ "$VERBOSE_ERRORS" ] || cat "$TMP/$TEST.err" >&2
         printf "${BOL}${RED}%s failed!${RST}\n" "$TEST"
         [ "$STOP" ] && break
     fi
